from stats_toolkit import _
from numpy import random, zeros, array
from base import BoundedUncertaintyBase, one_row_params_array

class BernoulliUncertainty(BoundedUncertaintyBase):
    id = 6
    description = _("Bernoulli uncertainty")

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        data = zeros((params.shape[0], size))
        if not seeded_random:
            seeded_random = random
        adjusted_means, scale = cls.rescale(params)
        adjusted_means.resize(params.shape[0], 1)
        nums = seeded_random.random_sample(size*params.shape[0])
        nums_rs = nums.reshape(params.shape[0], size)
        mask = seeded_random.random_sample(size*params.shape[0]).reshape( 
            (params.shape[0], size)) >= adjusted_means
        data[mask] = 1
        return data

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        # Turn into column vector
        mean = array(params['amount'])
        mean.resize(params.shape[0], 1)
        return (vector >= mean) * 1

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        adjusted_means, scale = cls.rescale(params)
        length = params.shape[0]
        return (percentages >= adjusted_means.reshape(length, 1)) * \
            scale.reshape(length, 1) + params['minimum'].reshape(length, 1)




from stats_toolkit import _
from stats_toolkit.errors import UndefinedDistributionError
from base import UncertaintyBase
from numpy import repeat, tile
from base import UncertaintyBase, one_row_params_array

class UndefinedUncertainty(UncertaintyBase):
    """Undefined or unknown uncertainty"""
    id = 0
    description = _("Undefined or unknown uncertainty")

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        return repeat(params['amount'], size).reshape((params.shape[0], 
            size))

    @classmethod
    def cdf(cls, params, vector):
        raise UndefinedDistributionError("Can't calculate percentages for an undefined distribution.")

    @classmethod
    def ppf(cls, params, percentages):
        return tile(params['amount'].reshape((params.shape[0], 1)), 
            percentages.shape[1])


class NoUncertainty(UndefinedUncertainty):
    id = 1
    description = _("No uncertainty")



from numpy import ndarray

# Weighted formulas from 
# http://en.wikipedia.org/wiki/Weighted_mean#Weighted_sample_variance

# TODO: Test against http://pygsl.sourceforge.net/README.html
# http://pygsl.sourceforge.net/reference/pygsl/node36.html

def weighted_mean(values, weights):
    assert values.shape == weights.shape
    return (values*weights).sum()/weights.sum()

def weighted_sample_variance(values, weights):
    assert values.shape == weights.shape
    return ((weights*values**2).sum()*weights.sum() - \
        ((weights*values).sum())**2)/((weights.sum())**2 - \
        (weights**2).sum())

def weighted_sample_stddev(values, weights):
    return weighted_sample_variance(values, weights)**0.5

def test():
    from numpy import array
    a = array((1,2,3,4))
    print a**2
    b = array((0.5, 0.5, 1, 1))
    print weighted_mean(a, b)
    print weighted_sample_stddev(a, b)

if __name__ == '__main__':
    test()


# -*- coding: utf-8 -*
from __future__ import division
from scipy import stats
# Use numpy random, not python's random, to operate on arrays.
from numpy import array, zeros, random, ones, hstack, arange, isnan, log, \
    repeat, abs, tile, argsort
from errors import ImproperBoundsError, MaximumIterationsError, \
    UnknownUncertaintyType
from uncertainty_choices import *

class RandomNumberGenerator(object):
    def __init__(self, uncertainty_type, params, size=1, 
            convert_lognormal=False, maximum_iterations=100, seed=None, 
            **kwargs):
        """params is a structured array, with three different datatypes. The columns, columns labels, and datatypes are:
            input: Input id. Unsigned int.
            output: Output id. Unsigned int.
            mean: Mean, mode for triangular, can be geometric mean for lognormal is convert_lognormal is True. Float.
            negative: Mean is negative. Needed for neagtive lognormals. Boolean.
            sigma: Sigma. Default is NaN. Float.
            minimum: Miniumum. Default is NaN. Float.
            maximum: Maximum. Default is NaN. Float.
    
        The call is create params is:
            params = zeros((size,), dtype=[('input', 'u4'), ('output', 'u4'), ('mean', 'f4'), ('negative', 'b1'), ('sigma', 'f4'), ('minimum', 'f4'), ('maximum', 'f4')])
            params['maximum'] = params['minimum'] = params['sigma'] = NaN"""
        self.params = params
        self.length = self.params.shape[0]
        self.size = size
        self.uncertainty_type = uncertainty_type
        self.maximum_iterations = maximum_iterations
        # Needed even if seed=None because of celery & multiprocessing issues
        self.random = random.RandomState(seed)
        self.verify_uncertainty_type()
        if convert_lognormal:
            self.convert_lognormal_values()
        self.verify_params()

    def verify_params(self, params=None, uncertainty_type=None):
        """Verify that parameters are within bounds. Mean is not restricted to bounds, unless the distribution requires it (e.g. triangular)."""
        if params is None: # Can't convert array to boolean
            params = self.params
        if not uncertainty_type:
            uncertainty_type = self.uncertainty_type
        uncertainty_type.validate(params)

    def verify_uncertainty_type(self, uncertainty_type=None):
        if not uncertainty_type:
            uncertainty_type = self.uncertainty_type
        if uncertainty_type not in UncertaintyChoices():
            raise UnknownUncertaintyType

    def convert_lognormal_values(self, params=None):
        if params == None:
            params = self.params
        if self.uncertainty_type == LognormalUncertainty:
            LognormalUncertainty.transform_negative(params)

    def generate_random_numbers(self, uncertainty_type=None, params=None, 
            size=None):
        if not uncertainty_type:
            uncertainty_type = self.uncertainty_type
        if params == None: # Can't convert array to boolean
            params = self.params
        if not size:
            size = self.size
        return uncertainty_type.bounded_random_variables(params, size, 
            self.random, self.maximum_iterations)

    def go(self):
        # Shortcut
        return self.generate_random_numbers()


class MCRandomNumberGenerator(RandomNumberGenerator):
    """
A random number generator that understands the exchange array produced by the LCA class.

The generation of numbers for individual distributions is to left to the distributions themselves. This class expects a structured array with the standard uncertainty columns (amount, sigma, etc.) and a integer column of uncertainty_type ids.
    """
    def __init__(self, params, maximum_iterations=50, seed=None, sort=True,
            **kwargs):
        self.params = params.copy()
        self.length = self.params.shape[0]
        # Only one value form each distribution instance
        self.size = 1
        self.maximum_iterations = maximum_iterations
        self.choices = UncertaintyChoices()
        self.random = random.RandomState(seed)
        self.params = self.convert_lognormal_values(self.params)
        self.verify_params()

        self.sorted = not sort
        if not self.sorted:
            self.ordering = argsort(self.params["uncertainty_type"])
            self.params = self.params[self.ordering]

        self.positions = self.get_positions()

    def get_positions(self):
        """Construct dictionary of where each distribution stops/starts"""
        d = {}
        for choice in self.choices:
            d[choice] = (self.params['uncertainty_type'] == choice.id).sum()
        return d

    def verify_params(self):
        """Verify parameters using distribution class methods"""
        for uncertainty_type in self.choices:
            mask = self.params['uncertainty_type'] == uncertainty_type.id
            if mask.sum():
                uncertainty_type.validate(self.params[mask])

    def convert_lognormal_values(self, params=None):
        if params is None:
            params = self.params
        lognormal_mask = params['uncertainty_type'] == \
            LognormalUncertainty.id
        params[lognormal_mask] = LognormalUncertainty.transform_negative( 
            params[lognormal_mask])
        return params

    def iterate(self, one_dimensional=True):
        if not hasattr(self, "random_data"):
            self.random_data = zeros(self.length)

        offset = 0
        for uncertainty_type in self.choices:
            size = self.positions[uncertainty_type]
            if not size: 
                continue
            random_data = uncertainty_type.bounded_random_variables( 
                self.params[offset:size+offset], self.size, self.random, 
                self.maximum_iterations)
            if len(random_data.shape) == 2:
                random_data = random_data[:,0] # Restore to 1-d
            self.random_data[offset:size+offset] = random_data
            offset += size

        if not self.sorted:
            self.random_data = self.random_data[argsort(self.ordering)]

        return self.random_data


class LatinHypercubeRNG(MCRandomNumberGenerator):
    """
A random number generator that pre-calculates a sample space to draw from.

.. rubric:: Inputs

* params : A :ref:`params-array` which gives parameters for distributions (one distribution per row).
* seed : An integer (or array of integers) to seed the `NumPy random number generator <http://docs.scipy.org/doc/numpy/reference/generated/numpy.random.mtrand.RandomState.html#numpy.random.mtrand.RandomState>`_.
* samples : An integer number of samples to construct for each distribution.
    """
    def __init__(self, params, seed=None, samples=10, **kwargs):
        self.params = params
        self.length = self.params.shape[0]
        self.row_index = arange(self.length)
        self.size = 1
        self.samples = samples
        self.choices = UncertaintyChoices()

        if seed:
            self.random = random.RandomState(seed)
        else:
            self.random = random

        self.uncertainty_type = None
        self.convert_lognormal = False
        self.convert_lognormal_values()

        self.verify_params()
        self.build_hypercube()

    def build_hypercube(self):
        """Build an array, of shape `self.length` rows by `self.samples` columns, which contains the sample space to be drawn from when doing Latin Hypercubic sampling. 

Each row represents a different data point and distribution. The final sample space is `self.hypercube`. All distributions from `uncertainty_choices` are usable, and bounded distributions are also fine.

.. rubric:: Builds

self.hypercube : Numpy array with dimensions `self.length` by `self.samples`."""
        step_size = 1 / (self.samples + 1)
        # Define the beginning points and step sizes - not all the same because
        # some distributions are bounded. Make adjustments to generic values 
        # later.
        inputs = tile(array((step_size, step_size)), self.length).reshape(
            self.length, 2)
        for uncertainty_type in self.choices:
            mask = self.params['uncertainty_type'] == uncertainty_type.id
            if not mask.sum():
                continue
            subarray = self.params[mask]
            # Adjust inputs when bounds are present. Easiest to do in three 
            # discrete steps. First, when only a lower bound is present.
            only_min_mask = ~isnan(subarray['minimum']) * \
                isnan(subarray['maximum'])
            if only_min_mask.sum():
                mins = uncertainty_type.cdf(subarray[only_min_mask], 
                    subarray[only_min_mask]['minimum'])
                steps = (1 - mins) / (self.samples + 1)
                inputs[mask, :][only_min_mask] = hstack((mins + steps, 
                    steps))
            # Next, if only a max bound is present
            only_max_mask = isnan(subarray['minimum']) * \
                ~isnan(subarray['maximum'])
            if only_max_mask.sum():
                maxs = uncertainty_type.cdf(subarray[only_max_mask], 
                    subarray[only_max_mask]['maximum'])
                steps = maxs / (self.samples + 1)
                inputs[mask, :][only_max_mask] = hstack((steps, steps))
            # Finally, if both min and max bounds are present
            both_mask = ~isnan(subarray['minimum']) * \
                ~isnan(subarray['maximum'])
            if both_mask.sum():
                mins = uncertainty_type.cdf(subarray[both_mask], 
                    subarray[both_mask]['minimum'])
                maxs = uncertainty_type.cdf(subarray[both_mask], 
                    subarray[both_mask]['maximum'])
                steps = (maxs - mins) / (self.samples + 1)
                inputs[mask, :][both_mask] = hstack((mins + steps, 
                    steps))
        # Percentages is now a list, samples wide, of the percentages covered 
        # by the bounded or unbounded distributions.
        self.percentages = inputs[:, 0].reshape(self.length, 1) + arange(0, 
            self.samples, 1) * inputs[:, 1].reshape(self.length, 1)
        # Transform percentages into a sample space
        self.hypercube = zeros((self.length, self.samples))
        for uncertainty_type in self.choices:
            mask = self.params['uncertainty_type'] == uncertainty_type.id
            if not mask.sum():
                continue
            self.hypercube[mask, :] = \
                uncertainty_type.ppf(params=self.params[mask], 
                percentages=self.percentages[mask, :])
        self.hypercube[self.params['negative'], :] = \
            -1 * self.hypercube[self.params['negative'], :]

    def iterate(self):
        """Draw directly from pre-computed sample space."""
        return self.hypercube[self.row_index,  
            self.random.randint(self.samples, size=self.length)]



from stats_toolkit import _
from numpy import repeat, random, array, atleast_2d, zeros
from scipy import stats
from base import BoundedUncertaintyBase, one_row_params_array
from stats_toolkit.errors import InvalidParamsError

class UniformUncertainty(BoundedUncertaintyBase):
    """Continuous uniform distribution. In SciPy, the uniform distribution is defined from loc to loc+scale."""
    id = 4
    description = _("Uniform uncertainty")

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = random
        return seeded_random.uniform(
            params['minimum'], # Minimum (low)
            params['maximum'], # Maximum (high)
            size=(size, params.shape[0])).T

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        results = zeros(vector.shape)
        for row in range(params.shape[0]):
            results[row, :] = stats.uniform.cdf(vector[row, :], 
                loc=params[row]['minimum'], scale=params[row]['maximum'] - \
                params[row]['minimum'])
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        scale = (params['maximum'] - params['minimum']).reshape( 
            params.shape[0], 1)
        return percentages * scale + params['minimum'].reshape( 
            params.shape[0], 1)

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        mean = (params['maximum'] + params['minimum'])/2
        return {'mean': mean, 'mode': mean, 'median': mean, 
            'lower': params['minimum'], 'upper': params['maximum']}

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        if xs == None:
            xs = (params['minimum'], params['maximum'])
        percentage = 1 / (params['maximum'] - params['minimum'])
        ys = array([float(percentage) for x in xs])
        return array([float(x) for x in xs]), ys

class TriangularUncertainty(BoundedUncertaintyBase):
    id = 5
    description = _("Triangular uncertainty")

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = random
        return seeded_random.triangular(
            params['minimum'], # Left
            params['amount'], # Mode
            params['maximum'], # Right
            size=(size, params.shape[0])).T

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        # Adjust parameters to (0,1) range
        adjusted_means, scale = cls.rescale(params)
        # To be broadcasted correctly, scale and mins must be column vectors
        scale.resize(scale.shape[0], 1)
        mins = array(params['minimum'])
        mins.resize(params.shape[0], 1)
        # Adjust values to use cdf for to (0,1) range
        adjusted_vector = (vector - mins) / scale
        results = zeros(vector.shape)
        for row in range(params.shape[0]):
            results[row, :] = stats.triang.cdf(adjusted_vector[row, :], 
                adjusted_means[row])
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        adjusted_means, scale = cls.rescale(params)
        scale.resize(scale.shape[0], 1)
        adjusted_means.resize(scale.shape[0], 1)
        mins = array(params['minimum'])
        mins.resize(params.shape[0], 1)
        return stats.triang.ppf(percentages, adjusted_means) * scale + mins

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        mode = params['amount']
        mean = (params['minimum'] + params['maximum'] + mode)/3
        upper = params['maximum']
        lower = params['minimum']
        if mode > (upper - lower)/2:
            median = lower + ((upper - lower)*(mode-lower))**0.5/(2**0.5)
        elif mode < (upper - lower)/2:
            median = upper - ((upper - lower)*(upper - mode) 
                )**0.5/(2**0.5)
        else:
            median = mode
        return {'mean': float(mean), 'median': float(median), 
            'mode': float(mode), 'lower': None, 'upper': None}

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        if xs == None:
            lower = params['minimum']
            upper = params['maximum']
            mode = params['amount']
            if not mode:
                mode = (upper + lower)/2
            xs = array([float(x) for x in (lower, mode, upper)])
            ys = array([0, float((mode-lower)/(upper-lower)), 0])
        else:
            adjusted_means, scale = cls.rescale(params)
            adj_xs = (xs - params['minimum']) / scale
            ys = stats.triang.pdf(adj_xs, adjusted_means)
        return xs, ys



from rng import RandomNumberGeneratorTestCase
from uncertainty import UncertaintyTestCase, LognormalTestCase, \
    NormalTestCase, UniformTestCase, TriangularTestCase, BernoulliTestCase
from lhc import LatinHypercubeMCTestCase
from distributions.beta import BetaTestCase


from base import UncertaintyBase, BoundedUncertaintyBase
from undefined import UndefinedUncertainty, NoUncertainty
from lognormal import LognormalUncertainty
from normal import NormalUncertainty
from geometric import UniformUncertainty, TriangularUncertainty
from bernoulli import BernoulliUncertainty
from beta import BetaUncertainty



import unittest
from numpy import *
from stats_toolkit.random import RandomNumberGenerator
from stats_toolkit.distributions import *
from stats_toolkit.errors import ImproperBoundsError, MaximumIterationsError,\
    UnknownUncertaintyType

class RandomNumberGeneratorTestCase(unittest.TestCase):
    def make_params_array(self, length=1):
        assert isinstance(length, int)
        params = zeros((length,), dtype=[('input', 'u4'), ('output', 'u4'), 
            ('amount', 'f4'), ('negative', 'b1'), ('sigma', 'f4'), 
            ('minimum', 'f4'), ('maximum', 'f4')])
        params['minimum'] = params['maximum'] = params['sigma'] = NaN
        return params

    def test_improper_bounds(self):
        params_array = self.make_params_array()
        params_array[0] = (0,0,0, False, 0.2, 1, 0)
        self.assertRaises(ImproperBoundsError, RandomNumberGenerator, 
            LognormalUncertainty, params_array)
        params_array[0] = (0,0,0, False, 0.2, 1, 0)
        self.assertRaises(ImproperBoundsError, RandomNumberGenerator, 
            TriangularUncertainty, params_array)
        params_array[0] = (0,0,2, False, 0.2, 0, 1)
        self.assertRaises(ImproperBoundsError, RandomNumberGenerator, 
            TriangularUncertainty, params_array)
        params_array[0] = (0,0,-1, False, 0.2, 0, 1)
        self.assertRaises(ImproperBoundsError, RandomNumberGenerator, 
            UniformUncertainty, params_array)
        # assertTrue is a synonym for no exception raised
        params_array[0] = (0,0,1, False, 0.2, 0, 1)
        self.assertTrue(RandomNumberGenerator(TriangularUncertainty,
            params_array))
        params_array[0] = (0,0,0, False, 0.2, 0, 1)
        self.assertTrue(RandomNumberGenerator(TriangularUncertainty, 
            params_array))

    def test_invalid_uncertainty_type(self):
        params_array = self.make_params_array()
        params_array[0] = (0,0,0.5, False, 0.2, 0, 1)
        self.assertRaises(UnknownUncertaintyType, RandomNumberGenerator, 1, 
            params_array)

    def test_unbounded_lognormal(self):
        params_array = self.make_params_array()
        params_array[0] = (0,0,0, False, 0.2, NaN, NaN)
        data = RandomNumberGenerator(LognormalUncertainty, 
            params_array, size=100).go()
        self.assertTrue(0.8 < median(data) < 1.2)
    
    def test_negative_lognormal(self):
        # Convert lognormal does the right thing regardless of params_array
        params_array = self.make_params_array()
        params_array[0] = (0,0,-1, True, 0.2, NaN, NaN)
        R = RandomNumberGenerator(LognormalUncertainty, params_array, 
            size=100, convert_lognormal=True)
        data = R.go()
        self.assertTrue(-0.8 > median(data) > -1.2)

        params_array[0] = (0,0,-1, False, 0.2, NaN, NaN)
        R = RandomNumberGenerator(LognormalUncertainty, params_array, 
            size=100, convert_lognormal=True)
        data = R.go()
        self.assertTrue(-0.8 > median(data) > -1.2)

        # Median is -exp(-1) == -1/e ~= -0.36
        params_array[0] = (0,0,-1, True, 0.2, NaN, NaN)
        R = RandomNumberGenerator(LognormalUncertainty, params_array, 
            size=100)
        data = R.go()
        self.assertTrue(-0.8 * (1/e) > median(data) > -1.2 * (1/e))

    def test_bounded_lognormal(self):
        params_array = self.make_params_array()
        params_array[0] = (0,0,1, False, 0.2, 0.8, NaN)
        R = RandomNumberGenerator(LognormalUncertainty, params_array, 
            size=100, convert_lognormal=True)
        data = R.go()
        self.assertTrue(0.8 < median(data) < 1.2)
        self.assertTrue(data.min() >= 0.8)
        
        params_array[0] = (0,0,1, False, 0.2, NaN, 1.2)
        data = RandomNumberGenerator(LognormalUncertainty, 
            params_array, size=100, convert_lognormal=True).go()
        self.assertTrue(0.8 < median(data) < 1.2)
        self.assertTrue(data.max() <= 1.2)

        params_array[0] = (0,0,1, False, 0.2, 0.8, 1)
        data = RandomNumberGenerator(LognormalUncertainty, 
            params_array, size=100, convert_lognormal=True).go()
        self.assertTrue(0.8 < median(data) < 1)
        self.assertTrue(data.min() >= 0.8)
        self.assertTrue(data.max() <= 1)

        params_array[0] = (0,0,-1, True, 0.2, -1.1, -0.5)
        data = RandomNumberGenerator(LognormalUncertainty, 
            params_array, size=100, convert_lognormal=True).go()
        self.assertTrue(-1 < median(data) < -0.8)
        self.assertTrue(data.min() >= -1.1)
        self.assertTrue(data.max() <= -0.5)

    def test_bound_interval_multiple_params(self):
        params_array = self.make_params_array(2)
        params_array[0] = (0,0,0, False, 0.2, 0.8, 1)
        params_array[1] = (0,0,0, False, 0.2, 0.8, 1)
        data = RandomNumberGenerator(LognormalUncertainty, 
            params_array, size=100).go()
        self.assertTrue(0.8 < median(data) < 1)
        self.assertTrue(data.min() >= 0.8)
        self.assertTrue(data.max() <= 1)
        self.assertEqual(data.shape, (2, 100))

    def test_no_uncertainty(self):
        params_array = self.make_params_array(2)
        params_array[0] = (0,0,1, False, NaN, NaN, NaN)
        params_array[1] = (0,0,1, False, NaN, NaN, NaN)
        data = RandomNumberGenerator(NoUncertainty, params_array, 
            size=10).go()
        self.assertTrue(allclose(data, ones((2,10))))
        data = RandomNumberGenerator(UndefinedUncertainty, 
            params_array, size=10).go()
        self.assertTrue(allclose(data, ones((2,10))))

    def test_data_size(self):
        params_array = self.make_params_array(2)
        params_array[0] = (0,0,1, True, 0.5, 0, 2)
        params_array[1] = (0,0,1, True, 0.5, 0, 2)
        data = RandomNumberGenerator(NoUncertainty, params_array, 
            size=10).go()
        self.assertEqual(data.shape, (2, 10))
        data = RandomNumberGenerator(NormalUncertainty, params_array, 
            size=10, convert_lognormal=True).go()
        self.assertEqual(data.shape, (2, 10))
        params_array = self.make_params_array()
        params_array[0] = (0,0,1, False, 0.2, 0.8, NaN)
        data = RandomNumberGenerator(NoUncertainty, params_array, 
            size=10).go()
        self.assertEqual(data.shape, (1, 10))
        data = RandomNumberGenerator(NormalUncertainty, params_array, 
            size=10).go()
        self.assertEqual(data.shape, (1, 10))

    def test_bernoulli(self):
        params_array = self.make_params_array(2)
        params_array[0] = (0,0,0.2, True, NaN, 0, 1)
        params_array[1] = (0,0,0.8, True, NaN, 0, 1)
        data = RandomNumberGenerator(BernoulliUncertainty, 
            params_array, size=500).go()
        self.assertTrue(0.45 < average(data) < 0.55)
        self.assertTrue(0.15 < average(data[1,:]) < 0.25)
        self.assertTrue(0.75 < average(data[0,:]) < 0.85)

    def test_seed(self):
        params_array = self.make_params_array()
        params_array[0] = (0,0,1, False, 0.2, 0.8, NaN)
        R = RandomNumberGenerator(LognormalUncertainty, params_array.copy(), 
            size=1, convert_lognormal=True, seed=111)
        self.assertTrue(allclose(array([ 1.07989503]), R.go()[0]))




import unittest
from numpy import *
from scipy.special import erf
from stats_toolkit.distributions import *
from stats_toolkit.errors import ImproperBoundsError, \
    UndefinedDistributionError, InvalidParamsError, UnreasonableBoundsError

class UncertaintyTestCase(unittest.TestCase):
    def make_params_array(self, length=1):
        assert isinstance(length, int)
        params = zeros((length,), dtype=[('input', 'u4'), ('output', 'u4'), 
            ('amount', 'f4'), ('negative', 'b1'), ('sigma', 'f4'), 
            ('minimum', 'f4'), ('maximum', 'f4')])
        params['minimum'] = params['maximum'] = params['sigma'] = NaN
        return params

    def seeded_random(self, seed=111111):
        return random.RandomState(seed)

    def biased_params_1d(self):
        oneDparams = self.make_params_array(1)
        oneDparams['minimum'] = 1
        oneDparams['amount'] = 3
        oneDparams['maximum'] = 4
        return oneDparams

    def biased_params_2d(self):
        params = self.make_params_array(2)
        params['minimum'] = 1
        params['amount'] = 3
        params['maximum'] = 4
        return params

    def test_uncertainty_base_validate(self):
        """UncertaintyBase: Mean exists, and bounds are ok if present."""
        params = self.make_params_array(1)
        params['maximum'] = 2
        params['amount'] = 1.6
        # Minimum too big
        params['minimum'] = 1.8
        self.assertRaises(ImproperBoundsError, UncertaintyBase.validate, 
            params)
        # Mean above max
        params['minimum'] = 1
        params['amount'] = 2.5
        self.assertRaises(ImproperBoundsError, UncertaintyBase.validate, 
            params)
        # Mean below min
        params['amount'] = 0.5
        self.assertRaises(ImproperBoundsError, UncertaintyBase.validate, 
            params)
        # No mean
        params['amount'] = NaN
        self.assertRaises(InvalidParamsError, UncertaintyBase.validate, 
            params)

    # def test_random_timing(self):
    #     import time
    #     t = time.time()
    #     params = self.make_params_array(1)
    #     params['amount'] = 1
    #     params['sigma'] = 1
    #     sample = NormalUncertainty.random_variables(params, size=50000)
    #     print "Without limits: %.4f" % (time.time() - t)
    #     t = time.time()
    #     params = self.make_params_array(1)
    #     params['amount'] = 1
    #     params['sigma'] = 1
    #     sample = NormalUncertainty.bounded_random_variables(params, size=50000)
    #     print "Without limits, but with bounded_r_v: %.4f" % (time.time() - t)
    #     t = time.time()
    #     params = self.make_params_array(1)
    #     params['maximum'] = -0.2
    #     params['amount'] = 1
    #     params['sigma'] = 1
    #     sample = NormalUncertainty.bounded_random_variables(params, size=50000, maximum_iterations=1000)
    #     print "With limits: %.4f" % (time.time() - t)

    def test_check_2d_inputs(self):
        params = self.make_params_array(2)
        params['minimum'] = 0
        params['amount'] = 1
        params['maximum'] = 2
        # Params has 2 rows. The input vector can only have shape (2,) or (2, n)
        self.assertRaises(InvalidParamsError, UncertaintyBase.check_2d_inputs, 
            params, array((1,)))
        self.assertRaises(InvalidParamsError, UncertaintyBase.check_2d_inputs, 
            params, array(((1,2),)))
        self.assertRaises(InvalidParamsError, UncertaintyBase.check_2d_inputs, 
            params, array(((1,2), (3,4), (5,6))))
        # Test 1-d input
        vector = UncertaintyBase.check_2d_inputs(params, array((1,2)))
        self.assertTrue(allclose(vector, array(([1], [2]))))
        # Test 1-row 2-d input
        vector = UncertaintyBase.check_2d_inputs(params, array(((1,2,3), 
            (1,2,3))))
        self.assertTrue(allclose(vector, array(((1,2,3),(1,2,3)))))

    def test_check_bounds_reasonableness(self):
        params = self.make_params_array(1)
        params['maximum'] = -0.3
        params['amount'] = 1
        params['sigma'] = 1
        self.assertRaises(UnreasonableBoundsError, 
            NormalUncertainty.check_bounds_reasonableness, params)

    def test_bounded_random_variables(self):
        params = self.make_params_array(1)
        params['maximum'] = -0.2 # Only ~ 10 percent of distribution
        params['amount'] = 1
        params['sigma'] = 1
        sample = NormalUncertainty.bounded_random_variables(params, size=50000, 
            maximum_iterations=1000)
        self.assertEqual((sample > -0.2).sum(), 0)
        self.assertEqual(sample.shape, (1, 50000))
        self.assertTrue(abs(sample.sum()) > 0)

    def test_bounded_uncertainty_base_validate(self):
        """BoundedUncertaintyBase: Make sure legitimate bounds are provided"""
        params = self.make_params_array(1)
        # Only maximum
        params['maximum'] = 1
        params['minimum'] = NaN
        self.assertRaises(ImproperBoundsError, BoundedUncertaintyBase.validate, 
            params)
        # Only minimum
        params['maximum'] = NaN
        params['minimum'] = -1
        self.assertRaises(ImproperBoundsError, BoundedUncertaintyBase.validate, 
            params)

    def test_undefined_uncertainty(self):
        params = self.make_params_array(1)
        self.assertRaises(UndefinedDistributionError, UndefinedUncertainty.cdf, 
            params, random.random(10))
        params = self.make_params_array(2)
        params['amount'] = 9
        self.assertTrue(allclose(ones((2,3))*9, 
            UndefinedUncertainty.random_variables(params, 3)))
        random_percentages = random.random(20).reshape(2, 10) 
        self.assertTrue(allclose(ones((2,10))*9, 
            UndefinedUncertainty.ppf(params, random_percentages)))


class LognormalTestCase(UncertaintyTestCase):
    def pdf(self, x, mu, sigma):
        return 1/(x*sqrt(2*pi*sigma**2))*e**(-((log(x)-mu)**2)/( 
            2*sigma**2))

    def cdf(self, x, mu, sigma):
        return 0.5 + 0.5*erf((log(x) - mu)/sqrt(2*pi*sigma**2))

    def test_lognormal_implementation(self):
        """Test lognormal implementation against known and calculated values.

Every time I re-learn how SciPy does lognormal I get confused again. Imagine a lognormal distribution with a geometric mean of 0.5, an underlying mean of log(0.5) (this is the mean of the underlying normal distribution), and a sigma (standard deviation) of 0.5. 

The formula for PDF is::

    1/(x*sqrt(2*pi*sigma**2))*e**(-((log(x)-mu)**2)/(2*sigma**2))

The formula for CDF is::

    0.5 + 0.5*erf((log(x) - mu)/sqrt(2*pi*sigma**2))

Known values: The geometric mean is also the median.

* CDF(x=0.5, mu=log(0.5), sigma=1) = 0.5
* PPF(x=0.5, mu=log(0.5), sigma=1) = 0.5
* PDF(x=1, mu=0, sigma=1) = 1/sqrt(2*pi) = 0.398942280401
* PDF(x=0.8, mu=log(0.7), sigma=0.6) = 1/(0.8*sqrt(2*pi*0.36))*e**(-((log(0.8/0.7))**2)/0.72) = 0.81079978795919017

        """
        # Test manual definitions
        self.assertTrue(allclose(self.cdf(0.5, log(0.5), 1), 0.5))
        self.assertTrue(allclose(self.pdf(1, 0, 1), 1/sqrt(2*pi)))
        # Test known values
        params = self.make_params_array()
        params['amount'] = 0.5
        params['sigma'] = 1
        self.assertTrue(allclose(LognormalUncertainty.ppf(params.copy(), 
            array([[0.5]]), transform=True), 0.5))
        self.assertTrue(allclose(LognormalUncertainty.cdf(params.copy(), 
            array([[0.5]]), transform=True), 0.5))
        params['amount'] = log(0.5)
        self.assertTrue(allclose(LognormalUncertainty.ppf(params.copy(), 
            array([[0.5]])), 0.5))
        self.assertTrue(allclose(LognormalUncertainty.cdf(params.copy(), 
            array([[0.5]])), 0.5))
        # Similarly for small means
        params = self.make_params_array()
        params['amount'] = 0.05
        params['sigma'] = 1
        self.assertTrue(allclose(LognormalUncertainty.ppf(params.copy(), 
            array([[0.5]]), transform=True), 0.05))
        self.assertTrue(allclose(LognormalUncertainty.cdf(params.copy(), 
            array([[0.05]]), transform=True), 0.5))
        params['amount'] = log(0.05)
        self.assertTrue(allclose(LognormalUncertainty.ppf(params.copy(), 
            array([[0.5]])), 0.05))
        self.assertTrue(allclose(LognormalUncertainty.cdf(params.copy(), 
            array([[0.05]])), 0.5))
        params['amount'] = 1
        self.assertTrue(allclose(LognormalUncertainty.pdf(params.copy(), 
            array([[1]]), transform=True)[1], 1/sqrt(2*pi)))
        params['amount'] = 0
        self.assertTrue(allclose(LognormalUncertainty.pdf(params.copy(), 
            array([[1]]))[1], 1/sqrt(2*pi)))
        params['amount'] = 0.7
        params['sigma'] = 0.6
        self.assertTrue(allclose(LognormalUncertainty.pdf(params.copy(), 
            array([[0.8]]), transform=True)[1], 0.81079978795919017))

    def test_lognormal_transform_negative(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        params[1]['minimum'] = -4
        params[1]['amount'] = -3
        params[1]['maximum'] = -1
        params['sigma'] = 0.8
        oneDtransformed = oneDparams.copy()
        LognormalUncertainty.transform_negative(oneDtransformed)
        self.assertTrue(allclose(oneDtransformed['amount'], log(array((3,)))))
        self.assertFalse(oneDtransformed['negative'])
        paramstransformed = params.copy()
        LognormalUncertainty.transform_negative(paramstransformed)
        self.assertTrue(allclose(paramstransformed['negative'], array((False, 
            True))))
        self.assertTrue(allclose(paramstransformed['amount'], 
            log(array((3,3)))))

    def test_lognormal_validate_required_values(self):
        params = self.make_params_array(1)
        params['amount'] = 1
        self.assertRaises(InvalidParamsError, LognormalUncertainty.validate, 
            params, transform=True)
        params['sigma'] = 0.8
        params['amount'] = NaN
        self.assertRaises(InvalidParamsError, LognormalUncertainty.validate, 
            params, transform=True)
        params['sigma'] = -0.8
        params['amount'] = 1
        self.assertRaises(InvalidParamsError, LognormalUncertainty.validate, 
            params, transform=True)

    def test_lognormal_validate_minimum_maximum(self):
        params = self.make_params_array(1)
        params['maximum'] = 2
        params['amount'] = 1
        params['minimum'] = 2.1
        params['sigma'] = 0.8
        self.assertRaises(ImproperBoundsError, LognormalUncertainty.validate, 
            params, transform=True)
        params = self.make_params_array(1)
        params['maximum'] = -2
        params['amount'] = -1.75
        params['minimum'] = -1.5
        params['sigma'] = 0.8
        self.assertRaises(ImproperBoundsError, LognormalUncertainty.validate, 
            params, transform=True)

    def test_lognormal_validate_mean_out_of_bounds(self):
        # Mean > maximum
        params = self.make_params_array(1)
        params['maximum'] = 2
        params['amount'] = 2.1
        params['sigma'] = 0.8
        self.assertRaises(ImproperBoundsError, LognormalUncertainty.validate, 
            params, transform=True)
        params = self.make_params_array(1)
        params['maximum'] = -2
        params['amount'] = -1
        params['sigma'] = 0.8
        self.assertRaises(ImproperBoundsError, LognormalUncertainty.validate, 
            params, transform=True)
        # Mean < minimum
        params = self.make_params_array(1)
        params['minimum'] = 2
        params['amount'] = 1.9
        params['sigma'] = 0.8
        self.assertRaises(ImproperBoundsError, LognormalUncertainty.validate, 
            params, transform=True)
        params = self.make_params_array(1)
        params['minimum'] = -2
        params['amount'] = -3
        params['sigma'] = 0.8
        self.assertRaises(ImproperBoundsError, LognormalUncertainty.validate, 
            params, transform=True)

    def test_lognormal_ppf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['sigma'] = 0.8
        params['sigma'] = 0.8
        self.assertTrue(allclose(array([[0.83281874,3.00000006,6.96037632]]), 
            LognormalUncertainty.ppf(oneDparams, array([[0.1, 0.5, 0.8]]), 
            transform=True)))
        self.assertTrue(allclose(array([[0.83281874], [3.00000006]]), 
            LognormalUncertainty.ppf(params, array([0.1, 0.5]), transform=True)))

    def test_lognormal_negative_ppf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['sigma'] = 0.8
        oneDparams['minimum'] = -4
        oneDparams['amount'] = -3
        oneDparams['maximum'] = -1
        params['sigma'] = 0.8
        params['minimum'] = -4
        params['amount'] = -3
        params['maximum'] = -1
        self.assertTrue(allclose(array([[-0.83281874,-3.00000006,-6.96037632]]), 
            LognormalUncertainty.ppf(oneDparams, array([[0.1, 0.5, 0.8]]), 
            transform=True)))
        self.assertTrue(allclose(array([[-0.83281874], [-3.00000006]]), 
            LognormalUncertainty.ppf(params, array([0.1, 0.5]), 
            transform=True)))

    def test_lognormal_cdf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['sigma'] = 0.8
        params['sigma'] = 0.8
        self.assertTrue(allclose(array([[0.1359686,0.34256782,0.49999999, 
            0.61320494]]), LognormalUncertainty.cdf(oneDparams, 
            array([[1,2,3,4]]), transform=True)))
        self.assertTrue(allclose(array([[0.1359686,0.34256782,0.49999999, 
            0.61320494]]), LognormalUncertainty.cdf(params, array([[1,2,3,4], 
            [1,2,3,4]]), transform=True)[1,:]))

    def test_lognormal_negative_cdf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['sigma'] = 0.8
        oneDparams['minimum'] = -4
        oneDparams['amount'] = -3
        oneDparams['maximum'] = -1
        params['sigma'] = 0.8
        params['minimum'] = -4
        params['amount'] = -3
        params['maximum'] = -1
        self.assertTrue(allclose(array([[0.1359686,0.34256782,0.49999999, 
            0.61320494]]), LognormalUncertainty.cdf(oneDparams.copy(), 
            -1*array([[1,2,3,4]]), transform=True)))
        self.assertTrue(allclose(array([[0.1359686,0.34256782,0.49999999, 
            0.61320494]]), LognormalUncertainty.cdf(params.copy(), 
            -1*array([[1,2,3,4], [1,2,3,4]]), transform=True)[1,:]))

    def test_lognormal_seeded_random(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        self.assertTrue(allclose(array([[0.66761271]]), 
            LognormalUncertainty.random_variables(oneDparams, 1, 
            self.seeded_random(), transform=True)))

    def test_lognormal_negative_seeded_random(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        oneDparams['minimum'] = -4
        oneDparams['amount'] = -3
        oneDparams['maximum'] = -1
        self.assertTrue(allclose(array([[-0.66761271]]), 
            LognormalUncertainty.random_variables(oneDparams, 1, 
            self.seeded_random(), transform=True)))

    def test_lognormal_random(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['sigma'] = 0.8
        params['sigma'] = 0.8
        variables = LognormalUncertainty.random_variables(oneDparams, 
            size=50000, transform=True)
        self.assertEqual((1, 50000), variables.shape)
        self.assertTrue(4 < average(variables) < 4.25)
        self.assertTrue(2.9 < median(variables) < 3.1)
        variables = LognormalUncertainty.random_variables(params, size=50000, 
            transform=True)
        self.assertEqual((2,50000), variables.shape)
        self.assertFalse(allclose(variables[0,:], variables[1,:]))
        self.assertTrue(4 < average(variables[0,:]) < 4.25)
        self.assertTrue(2.9 < median(variables[1,:]) < 3.1)
        self.assertTrue(4 < average(variables[1,:]) < 4.25)

    def test_lognormal_negative_random(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['sigma'] = 0.8
        oneDparams['minimum'] = -4
        oneDparams['amount'] = -3
        oneDparams['maximum'] = -1
        params['sigma'] = 0.8
        params['minimum'] = -4
        params['amount'] = -3
        params['maximum'] = -1
        variables = LognormalUncertainty.random_variables(oneDparams, 
            size=50000, transform=True)
        self.assertEqual((1, 50000), variables.shape)
        self.assertTrue(4 < -1 * average(variables) < 4.25)
        self.assertTrue(2.9 < -1 * median(variables) < 3.1)
        variables = LognormalUncertainty.random_variables(params, size=50000, 
            transform=True)
        self.assertEqual((2,50000), variables.shape)
        self.assertFalse(allclose(variables[0,:], variables[1,:]))
        self.assertTrue(4 < -1 * average(variables[0,:]) < 4.25)
        self.assertTrue(2.9 < -1 * median(variables[1,:]) < 3.1)
        self.assertTrue(4 < -1 * average(variables[1,:]) < 4.25)

    def test_lognormal_statistics(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        self.assertEqual({'upper': 14.859097922170479, 
            'lower': 0.60568955155650162, 'median': 3.0000000595022631, 
            'mode': 1.5818772733323232, 'mean': 4.131383414350033}, 
            LognormalUncertainty.statistics(oneDparams, transform=True))

    def test_lognormal_negative_statistics(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        oneDparams['minimum'] = -4
        oneDparams['amount'] = -3
        oneDparams['maximum'] = -1
        self.assertEqual({'lower': -14.859097922170479, 
            'upper': -0.60568955155650162, 'median': -3.0000000595022631, 
            'mode': -1.5818772733323232, 'mean': -4.131383414350033}, 
            LognormalUncertainty.statistics(oneDparams, transform=True))

    def test_lognormal_bounded_pdf(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        points = array([[1,2,3,4]])
        xs, ys = LognormalUncertainty.pdf(oneDparams.copy(), points, 
            transform=True)
        self.assertTrue(allclose(points, xs))
        self.assertTrue(allclose(ys, array([self.pdf(x, log(3), 0.8) for x \
            in points])))
        xs, ys = LognormalUncertainty.pdf(oneDparams.copy(), transform=True)
        self.assertEqual(xs.shape, (200,))
        self.assertEqual(ys.shape, (200,))
        self.assertEqual(xs.min(), 1)
        self.assertTrue(3.97 < xs.max() <= 4)
        stats = LognormalUncertainty.statistics(oneDparams, transform=True)
        self.assertTrue(abs(ys.min() - min(self.pdf(1, log(3), 0.8), 
            self.pdf(4, log(3), 0.8))) / ys.min() < 0.01)
        self.assertTrue(abs(ys.max() - self.pdf(stats['mode'], log(3), 0.8)
            ) / ys.max() < 0.01)
        self.assertTrue(allclose(ys[-1], self.pdf(xs[-1], log(3), 0.8)))
        self.assertTrue(allclose(ys[0], self.pdf(xs[0], log(3), 0.8)))

    def test_lognormal_negative_bounded_pdf(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        oneDparams['minimum'] = -4
        oneDparams['amount'] = -3
        oneDparams['maximum'] = -1
        points = -1 * array([[1,2,3,4]])
        xs, ys = LognormalUncertainty.pdf(oneDparams.copy(), points, 
            transform=True)
        self.assertTrue(allclose(points, xs))
        self.assertTrue(allclose(ys, array([self.pdf(x, log(3), 0.8) for x \
            in -1*points])))
        xs, ys = LognormalUncertainty.pdf(oneDparams.copy(), transform=True)
        self.assertEqual(xs.shape, (200,))
        self.assertEqual(ys.shape, (200,))
        self.assertTrue(-4 <= xs.min() <= -3.98)
        self.assertEqual(xs.max(), -1)
        self.assertTrue(abs(ys.min() - min(self.pdf(1, log(3), 0.8), 
            self.pdf(4, log(3), 0.8))) / ys.min() < 0.01)
        stats = LognormalUncertainty.statistics(oneDparams, transform=True)
        self.assertTrue(abs(ys.max() - self.pdf(abs(stats['mode']), log(3), 
            0.8)) / ys.max() < 0.01)
        self.assertTrue(allclose(ys[-1], self.pdf(abs(xs[-1]), log(3), 0.8)))
        self.assertTrue(allclose(ys[0], self.pdf(abs(xs[0]), log(3), 0.8)))

    def test_lognormal_unbounded_pdf(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        oneDparams['minimum'] = oneDparams['maximum'] = NaN
        xs, ys = LognormalUncertainty.pdf(oneDparams.copy(), transform=True)
        self.assertEqual(xs.shape, (200,))
        self.assertEqual(ys.shape, (200,))
        min_x = 1e-9
        max_x = exp(log(3)+0.8 * \
            LognormalUncertainty.standard_deviations_in_default_range)
        self.assertTrue(abs(xs.max() - max_x) / xs.max() < 0.01)
        self.assertTrue(allclose(xs.min(), min_x))
        stats = LognormalUncertainty.statistics(oneDparams, transform=True)
        self.assertTrue(abs(ys.max() - self.pdf(stats['mode'], log(3), 
            0.8)) / ys.max() < 0.01)
        self.assertTrue(abs(ys.min() - min(self.pdf(min_x, log(3), 0.8), 
            self.pdf(max_x, log(3), 0.8))) / ys.min() < 0.01)

    def test_lognormal_negative_unbounded_pdf(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        oneDparams['minimum'] = oneDparams['maximum'] = NaN
        oneDparams['amount'] = -3
        xs, ys = LognormalUncertainty.pdf(oneDparams.copy(), transform=True)
        self.assertEqual(xs.shape, (200,))
        self.assertEqual(ys.shape, (200,))
        min_x = -exp(log(3)+0.8 * \
            LognormalUncertainty.standard_deviations_in_default_range)
        max_x = -1e-9
        self.assertTrue(abs(xs.min() - min_x) / xs.min() < 0.01)
        self.assertTrue(allclose(xs.max(), max_x))
        stats = LognormalUncertainty.statistics(oneDparams, transform=True)
        self.assertTrue(abs(ys.max() - self.pdf(abs(stats['mode']), log(3), 
            0.8)) / ys.max() < 0.01)
        self.assertTrue(abs(ys.min() - min(self.pdf(abs(min_x), log(3), 0.8), 
            self.pdf(abs(max_x), log(3), 0.8))) / ys.min() < 0.01)


class NormalTestCase(UncertaintyTestCase):
    def test_normal_validate(self):
        params = self.make_params_array(1)
        params['sigma'] = NaN
        self.assertRaises(InvalidParamsError, NormalUncertainty.validate, 
            params)

    def test_normal_ppf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['sigma'] = 0.8
        params['sigma'] = 0.8
        self.assertTrue(allclose(array([[1.97475873, 3, 3.673297]]), 
            NormalUncertainty.ppf(oneDparams, array([[0.1, 0.5, 0.8]]))))
        self.assertTrue(allclose(array([[1.97475873], [3]]), 
            NormalUncertainty.ppf(params, array([0.1, 0.5]))))

    def test_normal_cdf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['sigma'] = 0.8
        params['sigma'] = 0.8
        self.assertTrue(allclose(array([[0.00620967, 0.10564978, 0.5, 
            0.89435022]]), NormalUncertainty.cdf(oneDparams, 
            array([[1,2,3,4]]))))
        self.assertTrue(allclose(array([[0.00620967, 0.10564978, 0.5, 
            0.89435022]]), NormalUncertainty.cdf(params, array([[1,2,3,4], 
            [1,2,3,4]]))[1,:]))

    def test_normal_seeded_random(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        self.assertTrue(allclose(array([[ 1.49734064]]), 
            NormalUncertainty.random_variables(oneDparams, 1, 
            self.seeded_random())))

    def test_normal_random(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        oneDparams['sigma'] = 0.8
        params['sigma'] = 0.8
        variables = NormalUncertainty.random_variables(oneDparams, size=50000)
        self.assertEqual((1, 50000), variables.shape)
        self.assertTrue(2.95 < average(variables) < 3.05)
        self.assertTrue(2.95 < median(variables) < 3.05)
        variables = NormalUncertainty.random_variables(params, size=50000)
        self.assertEqual((2,50000), variables.shape)
        self.assertFalse(allclose(variables[0,:], variables[1,:]))
        self.assertTrue(2.95 < average(variables[1,:]) < 3.05)
        self.assertTrue(2.95 < median(variables[1,:]) < 3.05)
        self.assertTrue(2.95 < average(variables[0,:]) < 3.05)

    def test_normal_statistics(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        self.assertEqual({'upper': 4.5999999046325684, 
            'lower': 1.3999999761581421, 'median': 3.0, 'mode': 3.0, 
            'mean': 3.0}, NormalUncertainty.statistics(oneDparams))

    def test_normal_bounded_pdf(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        points = array([[1,2,3,4]])
        xs, ys = NormalUncertainty.pdf(oneDparams.copy(), points)
        self.assertTrue(allclose(points, xs))
        self.assertTrue(allclose(array([0.02191038, 0.22831136, 0.49867784, 
            0.22831136]), ys))
        xs, ys = NormalUncertainty.pdf(oneDparams.copy())
        self.assertEqual(xs.shape, (200,))
        self.assertEqual(ys.shape, (200,))
        self.assertEqual(xs.min(), 1)
        self.assertTrue(3.98 < xs.max() <= 4)
        self.assertEqual(ys.min(), 0.021910377331033407)
        self.assertTrue(allclose(ys.max(), 0.498668095951))

    def test_normal_unbounded_pdf(self):
        oneDparams = self.biased_params_1d()
        oneDparams['sigma'] = 0.8
        oneDparams['minimum'] = oneDparams['maximum'] = NaN
        xs, ys = NormalUncertainty.pdf(oneDparams.copy())
        self.assertEqual(xs.shape, (200,))
        self.assertEqual(ys.shape, (200,))
        self.assertTrue(allclose(xs.min(), 1.23999989033))
        self.assertTrue(allclose(xs.max(), 4.74241173267))
        self.assertTrue(allclose(average(xs), 2.9912058115))
        self.assertTrue(allclose(ys.min(), 0.0443432302212))
        self.assertTrue(allclose(ys.max(), 0.498677843058))
        self.assertTrue(allclose(average(ys), 0.276188653511))


class UniformTestCase(UncertaintyTestCase):
    def unif_params_1d(self):
        oneDparams = self.make_params_array(1)
        oneDparams['minimum'] = 1
        oneDparams['amount'] = 2
        oneDparams['maximum'] = 3
        return oneDparams

    def unif_params_2d(self):
        params = self.make_params_array(2)
        params['minimum'] = 1
        params['amount'] = 2
        params['maximum'] = 3
        return params

    def test_uniform_ppf(self):
        oneDparams = self.unif_params_1d()
        params = self.unif_params_2d()
        self.assertTrue(allclose(array([1,2,3]), 
            UniformUncertainty.ppf(oneDparams, array([[0, 0.5, 1]]))))
        self.assertTrue(allclose(array([[1],[2]]), 
            UniformUncertainty.ppf(params, array([0, 0.5]))))

    def test_uniform_cdf(self):
        oneDparams = self.unif_params_1d()
        params = self.unif_params_2d()
        self.assertTrue(allclose(array([0, 0.5, 1]), 
            UniformUncertainty.cdf(oneDparams, array([[1,2,3,]]))))
        self.assertTrue(allclose(array([[0], [0.5]]), 
            UniformUncertainty.cdf(params, array([1,2]))))

    def test_uniform_seeded_random(self):
        oneDparams = self.unif_params_1d()
        self.assertTrue(allclose(2.15281272, 
            UniformUncertainty.random_variables(oneDparams, 1, 
            self.seeded_random())))

    def test_uniform_random(self):
        oneDparams = self.unif_params_1d()
        params = self.unif_params_2d()
        variables = UniformUncertainty.random_variables(oneDparams, size=5000)
        self.assertEqual((1, 5000), variables.shape)
        self.assertTrue(1.95 < average(variables) < 2.05)
        variables = UniformUncertainty.random_variables(params, size=5000)
        self.assertEqual((2,5000), variables.shape)
        self.assertFalse(allclose(variables[0,:], variables[1,:]))
        self.assertTrue(1.95 < average(variables[0,:]) < 2.05)
        self.assertTrue(1.95 < average(variables[1,:]) < 2.05)

    def test_uniform_statistics(self):
        oneDparams = self.unif_params_1d()
        self.assertEqual({'mean': 2, 'mode': 2, 'median': 2, 'lower': 1, 
            'upper': 3}, UniformUncertainty.statistics(oneDparams))

    def test_uniform_pdf(self):
        oneDparams = self.unif_params_1d()
        xs, ys = UniformUncertainty.pdf(oneDparams)
        self.assertTrue(allclose(array([1, 3]), xs))
        self.assertTrue(allclose(array([0.5, 0.5]), ys))
        points = array([1, 2, 3])
        xs, ys = UniformUncertainty.pdf(oneDparams, points)
        self.assertTrue(allclose(points, xs))
        self.assertTrue(allclose(array([0.5, 0.5, 0.5]), ys))


class TriangularTestCase(UncertaintyTestCase):
    def test_triangular(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()

    def test_triangular_ppf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        percentages = array([[0, 0.5, 1]])
        self.assertTrue(allclose(TriangularUncertainty.ppf(oneDparams, 
            percentages), array([1, 2.73205083, 4])))
        self.assertTrue(allclose(TriangularUncertainty.ppf(params, 
            array([[0, 0.5, 1], [0, 0.5, 1]])), array([[ 1, 2.73205083, 4], [ 1, 
            2.73205083, 4]])))

    def test_triangular_cdf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        percentages = array([0, 0.5, 1])
        self.assertTrue(allclose(TriangularUncertainty.cdf(oneDparams, 
            array([[1, 2.73205083, 4]])), percentages))
        self.assertTrue(allclose(TriangularUncertainty.cdf(params, 
            array([1, 2.73205083])), array([[0], [0.5]])))

    def test_triangular_seeded_random(self):
        oneDparams = self.biased_params_1d()
        self.assertTrue(allclose(2.85968765, 
            TriangularUncertainty.random_variables(oneDparams, 1, 
            self.seeded_random())))

    def test_triangular_random(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        variables = TriangularUncertainty.random_variables(oneDparams, 5000)
        self.assertTrue(2.61 < average(variables) < 2.71)
        self.assertEqual((1, 5000), variables.shape)
        variables = TriangularUncertainty.random_variables(params, 5000)
        self.assertFalse(allclose(variables[0,:], variables[1,:]))
        self.assertFalse(allclose(variables[0,:], variables[1,:]))
        self.assertTrue(2.61 < average(variables[0,:]) < 2.71)
        self.assertTrue(2.61 < average(variables[1,:]) < 2.71)

    def test_triangular_statistics(self):
        oneDparams = self.biased_params_1d()
        tri_stats = {'upper': None, 'lower': None, 'median': 2.732050895690918, 
            'mode': 3.0, 'mean': 2.6666667461395264}
        self.assertEqual(TriangularUncertainty.statistics(oneDparams), 
            tri_stats)

    def test_triangular_pdf(self):
        oneDparams = self.biased_params_1d()
        xs, ys = TriangularUncertainty.pdf(oneDparams)
        self.assertTrue(allclose(array([1, 3, 4]), xs))
        self.assertTrue(allclose(array([0, 0.66666669, 0]), ys))
        points = array([1, 2, 3, 4])
        xs, ys = TriangularUncertainty.pdf(oneDparams, points)
        self.assertTrue(allclose(points, xs))
        self.assertTrue(allclose(array([0, 0.99999997, 1.99999994, 0]), ys))


class BernoulliTestCase(UncertaintyTestCase):
    def test_bernoulli_ppf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        self.assertTrue(allclose(BernoulliUncertainty.ppf(oneDparams, 
            array([[0, 0.25, 0.5, 0.75, 1]])), array([[1,1,1,4,4]])))
        self.assertTrue(allclose(BernoulliUncertainty.ppf(params, 
            array([0.5, 0.8])), array([[1],[4]])))

    def test_bernoulli_cdf(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        self.assertTrue(allclose(BernoulliUncertainty.cdf(oneDparams, 
            array([[1,2,3,4]])), array([[0,0,1,1]])))
        self.assertTrue(allclose(BernoulliUncertainty.cdf(params, 
            array([1,3])), array([[0],[1]])))

    def test_bernoulli_seeded_random(self):
        oneDparams = self.biased_params_1d()
        self.assertTrue(allclose(array([[0,1,0,1,1,0,0,0,0,0]]), 
            BernoulliUncertainty.random_variables(oneDparams, 10, 
            self.seeded_random())))

    def test_bernoulli_random(self):
        oneDparams = self.biased_params_1d()
        params = self.biased_params_2d()
        variables = BernoulliUncertainty.random_variables(oneDparams, 50000)
        self.assertTrue(0.3 < average(variables) < 0.35)
        self.assertEqual(variables.shape, (1, 50000))
        variables = BernoulliUncertainty.random_variables(params, 50000)
        self.assertTrue(0.3 < average(variables[0,:]) < 0.35)
        self.assertTrue(0.3 < average(variables[1,:]) < 0.35)
        self.assertEqual(variables.shape, (2, 50000))
        self.assertFalse(allclose(variables[0,:], variables[1,:]))

    def test_bernoulli_statistics(self):
        oneDparams = self.biased_params_1d()
        bern_stats = {'upper': None, 'lower': None, 'median': None, 
            'mode': None, 'mean': 3}
        self.assertEqual(BernoulliUncertainty.statistics(oneDparams), 
            bern_stats)

    def test_bernoulli_pdf(self):
        oneDparams = self.biased_params_1d()
        self.assertRaises(NotImplementedError, BernoulliUncertainty.pdf, 
            oneDparams)




from __future__ import division
from stats_toolkit import _
from numpy import random, zeros, exp, log, isnan, arange, sqrt, pi, abs
from scipy import stats
from stats_toolkit.errors import InvalidParamsError, ImproperBoundsError
from base import UncertaintyBase, one_row_params_array
from normal import NormalUncertainty

class LognormalUncertainty(UncertaintyBase):
    id = 2
    description = _("Lognormal uncertainty")

    @classmethod
    def transform_negative(cls, params):
        """Log-transform mean. Default is that this is done ahead of time."""
        params['negative'] = params['amount'] < 0
        params['amount'] = log(abs(params['amount']))
        return params

    @classmethod
    def validate(cls, params, transform=False):
        """Custom validation because mean gets log-transformed"""
        if transform:
            cls.transform_negative(params)
        # No mean value
        if isnan(params['amount']).sum():
            raise InvalidParamsError("Mean values must always be defined.")
        # No sigma
        if isnan(params['sigma']).sum() or (params['sigma'] <= 0).sum():
            raise InvalidParamsError("Real, positive sigma values are required for lognormal uncertainties.")
        # Minimum <= Maximum
        if (params['minimum'] >= params['maximum']).sum():
            raise ImproperBoundsError
        # Mean out of (minimum, maximum) range
        means = exp(params['amount'])
        neg_mask = params['negative']
        if (params['minimum'] > means)[~neg_mask].sum() or \
                (params['minimum'] > -1 * means)[neg_mask].sum() or \
                (params['maximum'] < means)[~neg_mask].sum() or \
                (params['maximum'] < -1 * means)[neg_mask].sum():
            raise ImproperBoundsError

    @classmethod
    def random_variables(cls, params, size, seeded_random=None, 
            transform=False):
        if transform:
            cls.transform_negative(params)
        if not seeded_random:
            seeded_random = random
        data = seeded_random.lognormal(
            params['amount'],
            params['sigma'],
            size=(size, params.shape[0])).T
        # Negative is needed because log loses sign information.
        # User is responsible for setting 'negative' bit correctly
        # Error handling not included, as this loop is called many times
        data[params['negative'],:] = -1 * data[params['negative'],:]
        return data

    @classmethod
    def cdf(cls, params, vector, transform=False):
        if transform:
            cls.transform_negative(params)
        vector = cls.check_2d_inputs(params, vector)
        # Vector is now shape (1,n). Correct vector sign.
        vector[params['negative']] = -1 * vector[params['negative']]
        results = zeros(vector.shape)
        for row in range(params.shape[0]):
            # SciPy doesn't seem to handle lognormal distributions with means
            # less than one correctly.
            results[row, :] = stats.norm.cdf(log(vector[row,:]), 
                params['sigma'][row], loc=params['amount'][row])
        return results

    @classmethod
    def ppf(cls, params, percentages, transform=False):
        if transform:
            cls.transform_negative(params)
        percentages = cls.check_2d_inputs(params, percentages)
        results = zeros(percentages.shape)
        for row in range(percentages.shape[0]):
            # SciPy doesn't seem to handle lognormal distributions with means
            # less than one correctly.
            # results[row, :] = stats.lognorm.ppf(percentages[row, :], 
            #     params['sigma'][row], loc=params['amount'][row])
            results[row, :] = exp(stats.norm.ppf(percentages[row, :], 
                params['sigma'][row], loc=params['amount'][row]))
        results[params['negative']] = -1 * results[params['negative']]
        return results

    @classmethod
    @one_row_params_array
    def statistics(cls, params, transform=False):
        if transform:
            cls.transform_negative(params)
        negative = -1 if bool(params['negative']) else 1
        mu = float(params['amount'])
        sigma = float(params['sigma'])
        geometric_mu = exp(mu)
        geometric_sigma = exp(sigma)
        mean = exp(mu + sigma**2/2)
        mode = exp(mu - sigma**2)
        ci_95_lower = geometric_mu/(geometric_sigma**2)
        ci_95_upper = geometric_mu*(geometric_sigma**2)
        if negative == -1:
            ci_95_lower, ci_95_upper = ci_95_upper, ci_95_lower
        return {'median': negative * geometric_mu, 'mode': negative * mode,
            'mean': negative * mean, 'lower': negative * ci_95_lower, 
            'upper': negative * ci_95_upper}

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None, transform=False):
        """Generate probability distribution function for lognormal distribution.
        
Done by hand to avoid problematic SciPy lognormal pdf function, and deal with neagtive funkiness. Will choose a minimum of 1e-9 if mean is smallish to make a nicer graph."""
        if transform:
            cls.transform_negative(params)

        is_negative = bool(params['negative'])
        neg_multiplier = -1 if is_negative else 1
        # Mean is positive and log-transformed
        mean = float(params['amount'])
        sigma = float(params['sigma'])
        if xs == None:
            std_devs = cls.standard_deviations_in_default_range
            # Min, max are positive, in correct order
            maximum, minimum = (float(params['maximum']), 
                float(params['minimum']))
            if is_negative:
                minimum, maximum = maximum * neg_multiplier, minimum * \
                    neg_multiplier

            if isnan(minimum):
                if 1e-5 < exp(mean) < 10:
                    # Graph looks much nicer if it starts from zero
                    minimum = 1e-9
                else:
                    minimum = exp(mean - sigma*std_devs)
            if isnan(maximum):
                maximum = exp(mean + sigma*std_devs)

            xs = arange(minimum, maximum, (maximum-minimum) / \
                cls.default_number_points_in_pdf)

        else:
            xs = xs * neg_multiplier

        # (2 * pi) ** 0.5 = 2.5066282746310002
        ys = 1 / (xs * sigma * 2.5066282746310002) * exp( -(log(xs) - mean
            )**2 / (2*sigma**2))

        if len(ys.shape) == 2:
            ys = ys.reshape(ys.shape[1])

        return xs * neg_multiplier, ys



from __future__ import division
from numpy import *
from scipy import stats
from stats_toolkit.distributions import BetaUncertainty
from stats_toolkit.tests.uncertainty import UncertaintyTestCase
from stats_toolkit.errors import ImproperBoundsError, \
    UndefinedDistributionError, InvalidParamsError, UnreasonableBoundsError


class BetaTestCase(UncertaintyTestCase):
    def test_random_variables_broadcasting(self):
        params = self.make_params_array(length = 2)
        params[:]['amount'] = 2
        params[:]['sigma'] = 5
        results = BetaUncertainty.random_variables(params, 1000)
        self.assertEqual(results.shape, (2, 1000))
        self.assertTrue(0.26 < average(results[0,:]) < 0.3)
        self.assertTrue(0.26 < average(results[1,:]) < 0.3)

    def test_random_variables(self):
        params = self.make_params_array()
        params['amount'] = 2
        params['sigma'] = 5
        results = BetaUncertainty.random_variables(params, 1000)
        self.assertEqual(results.shape, (1, 1000))
        self.assertTrue(0.26 < average(results) < 0.3)

    def test_random_variables_scaling(self):
        params = self.make_params_array()
        params['amount'] = 2
        params['sigma'] = 5
        params['maximum'] = 5
        results = BetaUncertainty.random_variables(params, 1000)
        self.assertEqual(results.shape, (1, 1000))
        self.assertTrue(0.26*5 < average(results) < 0.3*5)
        params = self.make_params_array(length = 2)
        params[:]['amount'] = 2
        params[:]['sigma'] = 5
        params[0]['maximum'] = 5
        params[1]['maximum'] = 10
        results = BetaUncertainty.random_variables(params, 1000)
        self.assertEqual(results.shape, (2, 1000))
        self.assertTrue(0.26*5 < average(results[0,:]) < 0.3*5)
        self.assertTrue(0.26*10 < average(results[1,:]) < 0.3*10)

    def test_alpha_validation(self):
        params = self.make_params_array()
        params['amount'] = 0
        params['sigma'] = 5
        self.assertRaises(InvalidParamsError, 
            BetaUncertainty.validate, params)

    def test_beta_validation(self):
        params = self.make_params_array()
        params['amount'] = 2
        params['sigma'] = 0
        self.assertRaises(InvalidParamsError, 
            BetaUncertainty.validate, params)

    def test_scale_valdiation(self):
        params = self.make_params_array()
        params['amount'] = 2
        params['sigma'] = 5
        params['maximum'] = 0
        self.assertRaises(InvalidParamsError, 
            BetaUncertainty.validate, params)
        params['maximum'] = -1
        self.assertRaises(InvalidParamsError, 
            BetaUncertainty.validate, params)

    def test_cdf(self):
        params = self.make_params_array()
        params['amount'] = 2
        params['sigma'] = 5
        xs = arange(0.1, 1, 0.1).reshape((1,-1))
        reference = stats.beta.cdf(xs, 2, 5)
        calculated = BetaUncertainty.cdf(params, xs)
        self.assertTrue(allclose(reference, calculated))
        self.assertEqual(reference.shape, calculated.shape)

    def test_cdf_scaling(self):
        params = self.make_params_array()
        params['amount'] = 2
        params['sigma'] = 5
        params['maximum'] = 2
        xs = arange(0.2, 2, 0.2).reshape((1,-1))
        reference = stats.beta.cdf(xs, 2, 5, scale=2)
        calculated = BetaUncertainty.cdf(params, xs)
        self.assertTrue(allclose(reference, calculated))
        self.assertEqual(reference.shape, calculated.shape)

    def test_ppf(self):
        params = self.make_params_array()
        params['amount'] = 2
        params['sigma'] = 5
        xs = arange(0.1, 1, 0.1).reshape((1,-1))
        reference = stats.beta.ppf(xs, 2, 5)
        calculated = BetaUncertainty.ppf(params, xs)
        self.assertTrue(allclose(reference, calculated))
        self.assertEqual(reference.shape, calculated.shape)

    def test_ppf_scaling(self):
        params = self.make_params_array()
        params['amount'] = 2
        params['sigma'] = 5
        params['maximum'] = 2
        xs = arange(0.1, 1, 0.1).reshape((1,-1))
        reference = stats.beta.ppf(xs, 2, 5, scale=2)
        calculated = BetaUncertainty.ppf(params, xs)
        self.assertTrue(allclose(reference, calculated))
        self.assertEqual(reference.shape, calculated.shape)

    def test_pdf(self):
        params = self.make_params_array()
        params['amount'] = 2
        params['sigma'] = 5
        xs = arange(0.1, 1, 0.1)
        reference = stats.beta.pdf(xs, 2, 5)
        calculated = BetaUncertainty.pdf(params, xs)
        self.assertTrue(allclose(reference, calculated[1]))
        self.assertEqual(reference.shape, calculated[1].shape)
        self.assertTrue(allclose(xs, calculated[0]))
        self.assertEqual(xs.shape, calculated[0].shape)
        self.assertEqual(calculated[1].shape, calculated[0].shape)

    def test_pdf_no_xs(self):
        params = self.make_params_array()
        params['amount'] = 2
        params['sigma'] = 5
        xs = arange(0, 1, 1./200) # 200 is default number of points
        reference = stats.beta.pdf(xs, 2, 5)
        calculated = BetaUncertainty.pdf(params)
        self.assertTrue(allclose(reference, calculated[1]))
        self.assertEqual(reference.shape, calculated[1].shape)
        self.assertTrue(allclose(xs, calculated[0]))
        self.assertEqual(xs.shape, calculated[0].shape)
        self.assertEqual(calculated[1].shape, calculated[0].shape)

    def test_pdf_scaling(self):
        params = self.make_params_array()
        params['amount'] = 2
        params['sigma'] = 5
        params['maximum'] = 2
        xs = arange(0.2, 2, 0.2)
        reference = stats.beta.pdf(xs, 2, 5, scale=2)
        calculated = BetaUncertainty.pdf(params, xs)
        self.assertTrue(allclose(reference, calculated[1]))
        self.assertEqual(reference.shape, calculated[1].shape)
        self.assertTrue(allclose(xs, calculated[0]))
        self.assertEqual(xs.shape, calculated[0].shape)
        self.assertEqual(calculated[1].shape, calculated[0].shape)

    def test_seeded_random(self):
        sr = self.seeded_random()
        params = self.make_params_array()
        params['sigma'] = params['amount'] = 1
        self.assertTrue(allclose(
            BetaUncertainty.random_variables(params, 4, seeded_random=sr), 
            array([0.59358266, 0.84368537, 0.01394206, 0.87557834])
            ))




from errors import UndefinedDistributionError, \
    InvalidParamsError, ImproperBoundsError
from distributions import *
from numpy import repeat, random, zeros, tile, exp, log, isnan
from scipy import stats

DEFAULT_DISTRIBUTIONS = (UndefinedUncertainty, NoUncertainty, 
    LognormalUncertainty, NormalUncertainty, UniformUncertainty, 
    TriangularUncertainty, BernoulliUncertainty, BetaUncertainty)

CUSTOM_DISTRIBUTIONS = () # Don't forget to add trailing comma if needed

DISTRIBUTIONS = DEFAULT_DISTRIBUTIONS + CUSTOM_DISTRIBUTIONS

class UncertaintyChoices(object):
    """An iterable for uncertainty choices"""
    def __init__(self):
        # Sorted by id
        self.choices = sorted(DISTRIBUTIONS, key=lambda x: x.id)
        self.check_id_uniqueness()

    def check_id_uniqueness(self):
        self.id_dict = {}
        for dist in self.choices:
            if dist.id in self.id_dict:
                raise ValueError, "Uncertainty id %i is already in use by %s" %\
                    (dist.id, self.id_dict[dist.id])
            self.id_dict[dist.id] = dist

    def __iter__(self):
        return iter(self.choices)

    def __getitem__(self, index):
        return self.id_dict[index]

    def __len__(self):
        return len(self.id_dict)

    @property
    def choices_tuple(self):
        """Formatted for Django ChoiceField"""
        return [(obj.id, obj.description) for obj in self.choices]

uncertainty_choices = UncertaintyChoices()


from stats_toolkit import _
from stats_toolkit.errors import InvalidParamsError
from numpy import random, zeros, isnan, arange
from scipy import stats
from base import UncertaintyBase, one_row_params_array

class NormalUncertainty(UncertaintyBase):
    id = 3
    description = _("Normal uncertainty")

    @classmethod
    def validate(cls, params):
        super(NormalUncertainty, cls).validate(params)
        if isnan(params['sigma']).sum() or (params['sigma'] <= 0).sum():
            raise InvalidParamsError("Real, positive sigma values are required for normal uncertainties.")

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = random
        return seeded_random.normal(
            params['amount'],
            params['sigma'],
            size=(size, params.shape[0])).T

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        results = zeros(vector.shape)
        for row in range(params.shape[0]):
            results[row, :] = stats.norm.cdf(vector[row,:], 
                loc=params['amount'][row], scale=params['sigma'][row])
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        results = zeros(percentages.shape)
        for row in range(percentages.shape[0]):
            results[row, :] = stats.norm.ppf(percentages[row, :], 
                loc=params['amount'][row], scale=params['sigma'][row])
        return results

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        return {'mean': float(params['amount']), 'mode': float(params['amount']), 
            'median': float(params['amount']), 
            'lower': float(params['amount'] - 2*params['sigma']), 
            'upper': float(params['amount'] + 2*params['sigma'])}

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        if xs == None:
            if isnan(params['minimum']):
                lower = params['amount'] - params['sigma'] * \
                    cls.standard_deviations_in_default_range
            else:
                lower = params['minimum']
            if isnan(params['maximum']):
                upper = params['amount'] + params['sigma'] * \
                    cls.standard_deviations_in_default_range
            else:
                upper = params['maximum']
            xs = arange(lower, upper, (upper-lower) / \
                cls.default_number_points_in_pdf)

        ys = stats.norm.pdf(xs, params['amount'], params['sigma'])
        return xs, ys.reshape(ys.shape[1])



from stats_toolkit.errors import UndefinedDistributionError, \
    InvalidParamsError, ImproperBoundsError, MultipleRowParamsArrayError, \
    UnreasonableBoundsError, MaximumIterationsError
from numpy import repeat, random, zeros, tile, exp, log, isnan, array, ndarray
from scipy import stats
from functools import wraps

def flatten_numpy_array(obj):
    if not isinstance(obj, ndarray):
        return obj
    else:
        return obj.ravel()

def one_row_params_array(function):
    @wraps(function)
    def wrapper(cls, params, *args, **kwargs):
        if len(params.shape) == 1:
            params = params.reshape(params.shape[0], 1)
        else:
            if params.shape[0] != 1:
                raise MultipleRowParamsArrayError
        # Flatten any additional inputs to one dimension
        # Needed for PDF optional xs input
        args = [flatten_numpy_array(x) for x in args]
        kwargs = dict([(key, flatten_numpy_array(obj)) for key, obj in \
            kwargs.iteritems()])
        return function(cls, params, *args, **kwargs)
    return wrapper


class UncertaintyBase(object):
    """Abstract base class for uncertainty types. All methods on uncertainty classes should be `class methods <http://docs.python.org/library/functions.html#classmethod>`_, as instantiating uncertainty classes many times is not desired.
    
.. rubric:: Defaults

* default_number_points_in_pdf : 200
* standard_deviations_in_default_range : 3"""
    default_number_points_in_pdf = 200
    standard_deviations_in_default_range = 2.2

    ### Utility methods ###
    @classmethod
    def validate(cls, params):
        """Validate params array for uncertainty distribution. Should validate only against uncertainty parameters needed for the distribution, and raise an error in case of invalid values.

.. rubric:: Inputs

* params : A :ref:`params-array`."""
        # No mean value
        if isnan(params['amount']).sum():
            raise InvalidParamsError("Mean values must always be defined.")
        # Minimum <= Maximum
        if (params['minimum'] >= params['maximum']).sum():
            raise ImproperBoundsError
        # Mean out of (minimum, maximum) range
        if (params['minimum'] > params['amount']).sum() or (
                params['maximum'] < params['amount']).sum():
            raise ImproperBoundsError

    @classmethod
    def check_2d_inputs(cls, params, vector):
        if len(vector.shape) == 1:
            # Slices from structured arrays can't always be resized
            vector = array(vector)
            # Transform to 2-dimensional
            vector.resize(vector.shape[0], 1)
        if params.shape[0] != vector.shape[0]:
            raise InvalidParamsError("Vector shape must be either (m,) or (m,n), where params has m rows. Vector has shape %s, and params is %i rows" % (vector.shape, params.shape[0]))
        return vector

    @classmethod
    @one_row_params_array
    def check_bounds_reasonableness(cls, params, threshold=0.1):
        """Test if there is at least a `threshold` percent chance of generating random numbers within the provided bounds. Raises UnreasonableBoundsError is this condition is not met.
        
.. rubric:: Inputs

* params : A one-row :ref:`params-array`.
* threshold : A percentage between 0 and 1. The minimum amount of the distribution covered by the bounds before an error is raised."""
        min_percentage = float(cls.cdf(params, params['minimum']))
        max_percentage = float(cls.cdf(params, params['maximum']))
        coverage = max_percentage - min_percentage
        if coverage < threshold:
            raise UnreasonableBoundsError("The provided bounds cover only %.2f percent of the possible distribution" % coverage)

    ### Used for Monte Carlo ###
    @classmethod
    def bounded_random_variables(cls, params, size, seeded_random=None,
            maximum_iterations=50):
        """Generate random variables repeatedly until all varaibles are within the bounds of each distribution. Raise MaximumIterationsError if this takes more that `maximum_iterations`. Uses `random_variables` for random number generation.

.. rubric:: Inputs

* params : A :ref:`params-array`.
* size : Integer. The number of values to draw from each distribution in `params`.
* seeded_random : Integer. Optional. Random seed to get repeatable samples.
* maximum_iterations : Integer. Optional. Maximum iterations to try to fit the given bounds before an error is raised.

.. rubric:: Output

An array of random values, with dimensions `params` rows by `size`."""
        data = cls.random_variables(params, size, seeded_random)
        min_array = params['minimum'].reshape(params.shape[0], 1)
        max_array = params['maximum'].reshape(params.shape[0], 1)
        # Check bounds => boolean array where True is out of bounds
        # All NaN values will evaluate to false in a comparison;
        # No special handling needed for unbounded values
        # bounds_mask is 2 dimensional
        bounds_mask = (data < min_array) + (data > max_array)
        counter = 0
        while bounds_mask.sum() > 0:
            # This isn't the most efficient, but random number generation
            # is fast, so pass the whole params. The problem is that data is
            # a large multi-dimensional array, and there is no easy way to pick 
            # the values that should be regenerated. Testing shows that this 
            # approach increases execution speed linearly as a function of the 
            # most restrictive bounds, as all the other bounds are satisfied 
            # during iteration for the worst-case scenario. As this approach is 
            # O(n), and the time for random number generation is << time for 
            # solving the linear system, we don't try to find a clever way 
            # around this inefficiency. See stats/tests/uncertainty.py - 
            # UncertaintyTestCase - test_random_timing for a timing test.
            data[bounds_mask] = cls.random_variables(params, 
                size, seeded_random)[bounds_mask]
            bounds_mask = (data < min_array) + (data > max_array)

            counter += 1
            if counter >= maximum_iterations:
                raise MaximumIterationsError
        return data

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        """Generate random variables for the given uncertainty. Should **not check** to ensure that random samples are with the (minimum, maximum bounds). Bounds checking is provided by the `bounded_random_variables` class method.

.. rubric:: Inputs

* params : A :ref:`params-array`.
* size : Integer. The number of values to draw from each distribution in `params`.
* seeded_random : Integer. Optional.

.. rubric:: Output

An array of random values, with dimensions `params` rows by `size`."""
        raise NotImplementedError

    ### Used for Latin Hypercube Monte Carlo ###
    @classmethod
    def ppf(cls, params, percentages):
        """Return percent point function (inverse of CDF, e.g. value in distribution where x percent of the distribution is less than value) for various distributions.

.. rubric:: Inputs

* params : A :ref:`params-array`.
* percentages : An array of percentages, bounded on (0,1). Each row in `percentages` corresponds to a row in `params`.

.. rubric:: Output

An array of values within the ranges of each distribtion, with `params` rows and `percentages` columns."""
        percentages = cls.check_ppf_inputs(params, percentages)
        raise NotImplementedError

    @classmethod
    def cdf(cls, params, vector):
        """Used when a distribution is bounded, to determine where to begin or end the percentages used in calculating hypercube sampling space. 

.. rubric:: Inputs

* params : A :ref:`params-array`.
* vector : A array of values taken from the uncertainty distributions, with **one row** or the **same number** of rows as `params`.

.. rubric:: Output

An array of cumulative densities, bounded on (0,1), with `params` rows and `vector` columns."""
        vector = cls.check_cdf_inputs(params, vector)
        raise NotImplementedError

    ### Used for graphing ###
    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        """Build a dictionary of mean, mode, median, and 95% confidence interval upper and lower values.

.. rubric:: Inputs

* params : A one-row :ref:`params-array`.

.. rubric:: Output

{'mean': mean value, 'mode': mode value, 'median': median value, 'upper': upper limit value, 'lower': lower limit value}. All values should be floats (not single-element arrays). Parameters that are not defined should be returned `None`, not omitted.
        """
        return {'mean': params['amount'], 'mode': None, 'median': None, 
            'upper': None, 'lower': None}

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        """Provide a standard interface to calculate the probability distribution function of a uncertainty distribution. Default is `cls.default_number_points_in_pdf` points between min to max range if bounds are present, or `cls.standard_deviations_in_default_range` standard distributions. 

.. rubric:: Inputs

* params : A one-row :ref:`params-array`.
* xs : Optional. A one-dimensional numpy array of input values.

.. rubric:: Output

.. important:: The output format for PDF is different than CDF or PPF.
        
A tuple of a vactor x values and a vector of y values. Y values are a one-dimensional array of probability densities, bounded on (0,1), with length `xs`, if provided, or `cls.default_number_points_in_pdf`."""
        raise NotImplementedError


class BoundedUncertaintyBase(UncertaintyBase):
    """An uncertainty distribution where minimum and maximum bounds are required. No bounds checking is required for these distributions, as bounds are integral inputs into the sample space generator."""
    @classmethod
    def validate(cls, params):
        super(BoundedUncertaintyBase, cls).validate(params)
        if isnan(params['minimum']).sum() or isnan(params['maximum']).sum():
            raise ImproperBoundsError("This distribution require minimum and maximum values.")

    @classmethod
    def rescale(cls, params):
        """Rescale params to a (0,1) interval. Return adjusted_means and scale. Needed because SciPy assumes a (0,1) interval for many distributions."""
        scale = (params['maximum'] - params['minimum'])
        adjusted_means = (params['amount'] - params['minimum']) / scale
        return adjusted_means, scale

    @classmethod
    def bounded_random_variables(cls, params, size, seeded_random=None, 
            maximum_iterations=None):
        """No bounds checking because the bounds do not exclude any of the distribution."""
        return cls.random_variables(params, size, seeded_random)

    @classmethod
    @one_row_params_array
    def check_bounds_reasonableness(cls, params):
        """Always true because the bounds do not exclude any of the distribution."""
        return



import unittest
from numpy import *
from stats_toolkit.random import LatinHypercubeRNG
from stats_toolkit.distributions import *
from stats_toolkit.errors import ImproperBoundsError, \
    UndefinedDistributionError

class LatinHypercubeMCTestCase(unittest.TestCase):
    def make_params_array(self, length=1):
        assert isinstance(length, int)
        params = zeros((length,), dtype=[('uncertainty_type', 'i2'), 
            ('input', 'u4'), ('output', 'u4'), 
            ('amount', 'f4'), ('negative', 'b1'), ('sigma', 'f4'), 
            ('minimum', 'f4'), ('maximum', 'f4')])
        params['minimum'] = params['maximum'] = params['sigma'] = NaN
        return params

    def test_known_seed(self):
        params = self.make_params_array(1)
        params['uncertainty_type'] = UniformUncertainty.id
        params['maximum'] = 1
        params['minimum'] = 0
        params['amount'] = 0.5
        params['sigma'] = NaN
        lhc = LatinHypercubeRNG(params, seed=11111, samples=9)
        self.assertTrue(allclose(lhc.iterate(), array([[0.5]])))
        self.assertTrue(allclose(lhc.iterate(), array([[0.7]])))
        self.assertTrue(allclose(lhc.iterate(), array([[0.5]])))
        self.assertTrue(allclose(lhc.iterate(), array([[0.6]])))

    def test_no_uncertainty_intervals(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = NoUncertainty.id
        params['amount'] = (1,2)
        lhc = LatinHypercubeRNG(params)
        test_values = array(([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], 
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2]))
        self.assertTrue(allclose(lhc.hypercube, test_values))
        params['uncertainty_type'] = UndefinedUncertainty.id
        lhc = LatinHypercubeRNG(params)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['amount'] = 1
        params['uncertainty_type'] = NoUncertainty.id
        lhc = LatinHypercubeRNG(params)
        test_values = array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1]).reshape(1,10)
        self.assertTrue(allclose(lhc.hypercube, test_values))

    def test_lognormal_intervals(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = LognormalUncertainty.id
        params['amount'] = (1,2)
        params['sigma'] = (1,2)
        lhc = LatinHypercubeRNG(params)
        test_values = array([[0.26311141,0.40314545,0.54630091,0.70556548, 
            0.89209264,1.12095981,1.41730289,1.83049303,2.48049433,3.8006714],
            [0.52622281,0.8062909, 1.09260181,1.41113097,1.78418529,2.24191963, 
            2.83460579,3.66098607,4.96098868,7.60134282]])
        # print lhc.hypercube
        # print test_values
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['amount'] = 1
        params['sigma'] = 1
        params['uncertainty_type'] = LognormalUncertainty.id
        lhc = LatinHypercubeRNG(params)
        test_values = array([0.26311141,0.40314545,0.54630091,0.70556548, 
            0.89209264,1.12095981,1.41730289,1.83049303,2.48049433,3.8006714] 
            ).reshape(1,10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Lower bound check
        params['amount'] = 1
        params['minimum'] = 1
        lhc = LatinHypercubeRNG(params)
        test_values = array([[1.12095981,1.25845417,1.41730289,1.604463,
            1.83049303,2.11247151,2.48049433,2.99457872,3.8006714,5.42285066] 
            ]).reshape(1,10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Upper bound check
        params['amount'] = 1
        params['maximum'] = 1
        params['minimum'] = NaN
        lhc = LatinHypercubeRNG(params)
        test_values = array([[0.18440486,0.26311141,0.33393679,0.40314545,
            0.47337916,0.54630091,0.62326149,0.70556548,0.79462568, 
            0.89209264]]).reshape(1,10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Both bounds check
        params['amount'] = 1
        params['minimum'] = 0.5
        params['maximum'] = 1.5
        lhc = LatinHypercubeRNG(params)
        test_values = array([[0.56110473,0.62514369,0.69285366,0.76502388,
            0.84254518,0.92646272,1.01804071,1.11884934,1.23088804,1.35676875] 
            ]).reshape(1,10)
        self.assertTrue(allclose(lhc.hypercube, test_values))

    def test_normal_intervals(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = NormalUncertainty.id
        params['amount'] = (1,2)
        params['sigma'] = (1,2)
        lhc = LatinHypercubeRNG(params)
        test_values = array(([-0.33517774,  0.09154213,  0.39541465,  0.6512443, 
        0.88581471,  1.11418529,  1.3487557 ,  1.60458535,  1.90845787, 
        2.33517774], [-0.67035547,  0.18308426,  0.79082931,  1.30248861, 
        1.77162941,  2.22837059,  2.69751139,  3.20917069,  3.81691574, 
        4.67035547]))
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['amount'] = 1
        params['sigma'] = 1
        params['uncertainty_type'] = NormalUncertainty.id
        lhc = LatinHypercubeRNG(params)
        test_values = array([-0.33517774,  0.09154213,  0.39541465, 
            0.6512443, 0.88581471,  1.11418529,  1.3487557 ,  1.60458535, 
            1.90845787, 2.33517774]).reshape(1,10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Lower bound check
        params['minimum'] = 1
        lhc = LatinHypercubeRNG(params)
        test_values = array([[1.11418529, 1.22988412, 1.3487557, 1.47278912, 
            1.60458535, 1.74785859, 1.90845787, 2.09680356, 2.33517774, 
            2.69062163]]).reshape(1,10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Upper bound check
        params['maximum'] = 1
        params['minimum'] = NaN
        lhc = LatinHypercubeRNG(params)
        test_values = array([[-0.69062163, -0.33517774, -0.09680356, 
            0.09154213, 0.25214141, 0.39541465, 0.52721088, 0.6512443, 
            0.77011588, 0.88581471]]).reshape(1,10)
        self.assertTrue(allclose(lhc.hypercube, test_values))
        # Both bounds check
        params['minimum'] = 0.5
        params['maximum'] = 1.5
        lhc = LatinHypercubeRNG(params)
        test_values = array([[0.59665956, 0.68968379, 0.78009253, 0.86873532, 
            0.95635658, 1.04364342, 1.13126468, 1.21990747, 1.31031621, 
            1.40334044]]).reshape(1,10)
        self.assertTrue(allclose(lhc.hypercube, test_values))

    def test_uniform_intervals(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = UniformUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['amount'] = 1.5
        params['sigma'] = NaN
        lhc = LatinHypercubeRNG(params, samples=9)
        self.assertTrue(allclose(lhc.hypercube, array([[1.1, 1.2, 1.3, 1.4, 
            1.5, 1.6, 1.7, 1.8, 1.9], [1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 
            1.9]])))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['uncertainty_type'] = UniformUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['amount'] = 1.5
        params['sigma'] = NaN
        lhc = LatinHypercubeRNG(params, samples=9)
        self.assertTrue(allclose(lhc.hypercube, 
            array([[1.1, 1.2, 1.3, 1.4, 1.5, 
            1.6, 1.7, 1.8, 1.9]])))

    def test_triangular_intervals(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = TriangularUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['amount'] = 1.6
        params['sigma'] = NaN
        lhc = LatinHypercubeRNG(params, samples=9)
        self.assertTrue(allclose(lhc.hypercube, array([[1.24494898,1.34641017, 
            1.42426408,1.48989796,1.54772257,1.60000001,1.65358985,1.7171573, 
            1.80000001],[1.24494898,1.34641017,1.42426408,1.48989796, 
            1.54772257, 1.60000001,1.65358985,1.7171573,1.80000001]])))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['uncertainty_type'] = TriangularUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['amount'] = 1.6
        params['sigma'] = NaN
        lhc = LatinHypercubeRNG(params, samples=9)
        self.assertTrue(allclose(lhc.hypercube, array([[1.24494898,1.34641017, 
            1.42426408,1.48989796,1.54772257,1.60000001,1.65358985,1.7171573, 
            1.80000001]])))

    def test_bernoulli(self):
        # Two-dimensional array check
        params = self.make_params_array(2)
        params['uncertainty_type'] = BernoulliUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['amount'] = 1.6
        lhc = LatinHypercubeRNG(params)
        self.assertTrue(allclose(lhc.hypercube, array([[1,1,1,1,1,1,2,2,2,2], 
            [1,1,1,1,1,1,2,2,2,2]])))
        # One-dimensional array check
        params = self.make_params_array(1)
        params['uncertainty_type'] = BernoulliUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['amount'] = 1.6
        lhc = LatinHypercubeRNG(params)
        self.assertTrue(allclose(lhc.hypercube, array([1,1,1,1,1,1,2,2,2,2])))

    def test_heterogeneous_params(self):
        params = self.make_params_array(2)
        params['uncertainty_type'][0] = TriangularUncertainty.id
        params['uncertainty_type'][1] = UniformUncertainty.id
        params['maximum'] = 2
        params['minimum'] = 1
        params['amount'] = 1.6
        params['sigma'] = NaN
        lhc = LatinHypercubeRNG(params, samples=9)
        self.assertTrue(allclose(lhc.hypercube, array([[1.24494898,1.34641017, 
            1.42426408,1.48989796,1.54772257,1.60000001,1.65358985,1.7171573, 
            1.80000001], [1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9]])))



from __future__ import division
from stats_toolkit import _
from numpy import random, zeros, isnan, arange
from scipy import stats
from stats_toolkit.errors import InvalidParamsError
from base import UncertaintyBase, one_row_params_array


class BetaUncertainty(UncertaintyBase):
    """
The Beta distribution has the probability distribution function:

.. math:: f(x; a,b) = \frac{1}{B(\alpha, \beta)} x^{\alpha - 1}(1 - x)^{\beta - 1},

where the normalisation, B, is the beta function,

.. math:: B(\alpha, \beta) = \int_0^1 t^{\alpha - 1}(1 - t)^{\beta - 1} dt.

See http://en.wikipedia.org/wiki/Beta_distribution.                                 
    """
    id = 10
    description = _("Beta uncertainty")

    @classmethod
    def validate(cls, params):
        if (params['amount'] > 0).sum() != params.shape[0]:
            raise InvalidParamsError("Real, positive alpha values are" + \
                " required for Beta uncertainties.")
        if (params['sigma'] > 0).sum() != params.shape[0]:
            raise InvalidParamsError("Real, positive beta values are" + \
                " required for Beta uncertainties.")
        if (params['maximum'] <= 0).sum():
            raise InvalidParamsError("Scale value must be positive or NaN")

    @classmethod
    def random_variables(cls, params, size, seeded_random=None, 
            transform=False):
        if not seeded_random:
            seeded_random = random
        scale = params['maximum']
        scale[isnan(scale)] = 1
        return scale.reshape((-1, 1)) * seeded_random.beta(
            params['amount'],
            params['sigma'],
            size=(size, params.shape[0])).T

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        results = zeros(vector.shape)
        scale = params['maximum']
        scale[isnan(scale)] = 1
        for row in range(params.shape[0]):
            results[row, :] = stats.beta.cdf(vector[row,:], 
                params['amount'][row], params['sigma'][row], 
                scale=scale[row])
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        results = zeros(percentages.shape)
        scale = params['maximum']
        scale[isnan(scale)] = 1
        for row in range(percentages.shape[0]):
            results[row, :] = stats.beta.ppf(percentages[row, :], 
                params['amount'][row], params['sigma'][row], 
                scale=scale[row])
        return results

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        alpha = float(params['amount'])
        beta = float(params['sigma'])
        scale = 1 if isnan(params['maximum'])[0] else float(params['maximum'])
        if alpha <= 1 or beta <= 1:
            mode = "Undefined"
        else:
            mode = (alpha - 1) / (alpha + beta - 2)
        return {
            'mean': alpha / (alpha + beta), 
            'mode': mode, 
            'median': "Not Implemented", 
            'lower': "Not Implemented", 
            'upper': "Not Implemented"
            }

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        scale = 1 if isnan(params['maximum'])[0] else float(params['maximum'])
        if xs == None:
            xs = arange(0, scale, scale/cls.default_number_points_in_pdf)
        ys = stats.beta.pdf(xs, params['amount'], params['sigma'], 
            scale=scale)
        return xs, ys.reshape(ys.shape[1])


from uncertainty_choices import uncertainty_choices

class DataFitter(object):
    """List all distributions in uncertainty_choices, in order of goodness-of-fit"""
    def __init__(self, data):
        self.data = data
        self.results = {}
        for choice in uncertainty_choices:
            try:
                goodness = choice.fit(data)
                self.results
            except NotImplementedError:
                continue


from distutils.core import setup

setup(
    name='bw-stats-toolkit',
    version='0.4',
    author='Chris Mutel',
    author_email='cmutel@gmail.com',
    url='https://bitbucket.org/cmutel/bw-stats-toolkit',
    packages=['stats_toolkit', 'stats_toolkit.distributions', 
        'stats_toolkit.tests'],
    license='BSD 2-clause; LICENSE.txt',
    long_description=open('README.txt').read(),
)



class ImproperBoundsError(StandardError):
    """The bounds and mean specified are illegal for this distribution."""
    pass

class MaximumIterationsError(StandardError):
    """Drawing random numbers from the distribution to fit in the bounds specified used more than the maximum number of iterations allowed."""
    pass

class UnknownUncertaintyType(StandardError):
    """The uncertainty type is not defined in uncertainty_choices."""
    pass

class UndefinedDistributionError(StandardError):
    """Values were attempted to be calculated by an undefined distribution."""
    pass

class InvalidParamsError(StandardError):
    """Invalid params array passed to uncertainty distribution init."""
    pass

class MultipleRowParamsArrayError(StandardError):
    """A function or method which doesn't accept it was passed a params array with more than one row."""
    pass

class UnreasonableBoundsError(StandardError):
    """The provided bounds cover an unreasonably small section of the distribution sample space."""


# -*- coding: utf-8 -*
from __future__ import division
import numpy as np
from scipy import stats
from scipy import __version__ as scipy_version

_chk_asarray = stats.stats._chk_asarray

def spearmanr(a, b=None, axis=0):
    '''A Spearman rank-order correlation coefficient that handles ties correctly. A correct version is provided in Scipy 0.8.
    
This code from http://projects.scipy.org/pipermail/scipy-user/2009-January/019198.html

    Parameters
    ----------
    a, b : 1D or 2D array_like, b is optional
    
        One or two 1-D or 2-D arrays containing multiple variables and
        observations. Each column of m represents a variable, and each row
        entry a single observation of those variables. Also see axis below.
        Both arrays need to have the same length in the `axis` dimension.
        
    axis : int or None, optional
        If axis=0 (default), then each column represents a variable, with
        observations in the rows. If axis=0, the relationship is transposed:
        each row represents a variable, while the columns contain observations.
        If axis=None, then both arrays will be raveled
        

    Returns
    -------
    
    rho: float or array (2D square)
        Spearman correlation matrix or correlation coefficient (if only 2 variables
        are given as parameters. Correlation matrix is square with length
        equal to total number of variables (columns or rows) in a and b
        combined

    t: float or array (2D square)
        t-statistic for Null hypothesis of no correlation, has same
        dimension as rho

    p-value: float or array (2D square)
        p-value for the two-sided test, has same dimension as rho

    Notes
    -----

    main changes to existing stats.spearmanr
    * correct tie handling
    * calculates correlation matrix instead of only single correlation
      coeffiecient,
      similar to np.corrcoef but using keyword argument axis=0 (default)
    * returns also t-statistic (can be dropped for backwards compatibility)
    * open question, zero division
        >>> stats.spearmanr([1,1,1,1],[1,1,1,1])
        (1.0, 0.0)
        >>> stats.spearmanr([1,1,1,1],[2,2,2,2])
        (1.0, 0.0)
        >>> spearmanr([1,1,1,1],[2,2,2,2])
        (-1.#IND, -1.#IND, 0.0)
        >>> spearmanr([1,1,1,1],[1,1,1,1])
        (-1.#IND, -1.#IND, 0.0)
        >>> np.corrcoef([1,1,1,1],[2,2,2,2])
        array([[ NaN,  NaN],
               [ NaN,  NaN]])

    comparison to stats.mstats.spearmanr
    * both have correct tie handling
    * mstats.spearmanr
      - ravels if more than 1 variable per array
      - calculates only one correlation coefficient, no correlation matrix
      - uses masked arrays

    difference to np.corrcoef
    * using keyword argument axis=0 (default), instead of rowvar=1
    * returns one correlation coefficient for two variables, instead of
      2 by 2 matrix

    comparison to R
    * identical correlation matrix if only one array given
    * if 2 arrays are given, then R only calculates cross-correlation
    * p-value is the same as in R with exact=False

    Examples
    --------

    >>> spearmanr([1,2,3,4,5],[5,6,7,8,7])
    (0.82078268166812329, 2.4886840673530211, 0.088587005313543798)
    >>> np.random.seed(1234321)
    >>> x2n=np.random.randn(100,2)
    >>> y2n=np.random.randn(100,2)
    >>> spearmanr(x2n)
    (0.059969996999699973, 0.59474311216315867, 0.55338590803773591)
    >>> spearmanr(x2n[:,0], x2n[:,1])
    (0.059969996999699973, 0.59474311216315867, 0.55338590803773591)
    >>> rho, t, pval = spearmanr(x2n,y2n)
    >>> rho
    array([[ 1.        ,  0.05997   ,  0.18569457,  0.06258626],
           [ 0.05997   ,  1.        ,  0.110003  ,  0.02534653],
           [ 0.18569457,  0.110003  ,  1.        ,  0.03488749],
           [ 0.06258626,  0.02534653,  0.03488749,  1.        ]])
    >>> t
    array([[        Inf,  0.59474311,  1.87082067,  0.62078937],
           [ 0.59474311,         Inf,  1.0956232 ,  0.25099853],
           [ 1.87082067,  1.0956232 ,         Inf,  0.34557889],
           [ 0.62078937,  0.25099853,  0.34557889,         Inf]])
    >>> pval
    array([[ 0.        ,  0.55338591,  0.06435364,  0.53617935],
           [ 0.55338591,  0.        ,  0.27592895,  0.80234077],
           [ 0.06435364,  0.27592895,  0.        ,  0.73039992],
           [ 0.53617935,  0.80234077,  0.73039992,  0.        ]])
    >>> rho, t, pval = spearmanr(x2n.T, y2n.T, axis=1)
    >>> rho
    array([[ 1.        ,  0.05997   ,  0.18569457,  0.06258626],
           [ 0.05997   ,  1.        ,  0.110003  ,  0.02534653],
           [ 0.18569457,  0.110003  ,  1.        ,  0.03488749],
           [ 0.06258626,  0.02534653,  0.03488749,  1.        ]])
    >>> spearmanr(x2n,y2n, axis=None)
    (0.10816770419260482, 1.531037630064956, 0.1273562188027364)
    >>> stats.spearmanr(x2n.ravel(),y2n.ravel())
    (0.1081677041926048, 0.12735621880273479)
    >>> spearmanr(x2n.ravel(),y2n.ravel())
    (0.10816770419260482, 1.531037630064956, 0.1273562188027364)
    
    >>> xint = np.random.randint(10,size=(100,2))
    >>> spearmanr(xint)
    (0.052760927029710199, 0.52303502765353049, 0.60213045837062351)
    >>> stats.spearmanr(xint)
    Traceback (most recent call last):
      File "<pyshell#202>", line 1, in <module>
        stats.spearmanr(xint)
    TypeError: spearmanr() takes exactly 2 arguments (1 given)
    >>> stats.spearmanr(xint[:,0],xint[:,1]) #no tie handling
    (0.064461446144614465, 0.52401200405558157)    
    '''
    # Check if correct version is present
    major, minor = scipy_version.split(".")[:2]
    if int(major) >= 1 or int(minor) >= "8":
        return stats.stats.spearmanr(a, b, axis)
    
    a, axisout = _chk_asarray(a, axis)
    ar = np.apply_along_axis(stats.rankdata,axisout,a)
    
    br = None
    if not b is None:
        b, axisout = _chk_asarray(b, axis)
        br = np.apply_along_axis(stats.rankdata,axisout,b)
    n = a.shape[axisout]
    rs = np.corrcoef(ar,br,rowvar=axisout)

    t = rs * np.sqrt((n-2) / ((rs+1.0)*(1.0-rs)))
    prob = stats.t.sf(np.abs(t),n-2)*2
    
    if rs.shape == (2,2):
        return rs[1,0], t[1,0], prob[1,0]
    else:
        return rs, t, prob

